# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
#
# Code generated by aaz-dev-tools
# --------------------------------------------------------------------------------------------

# pylint: skip-file
# flake8: noqa

from azure.cli.core.aaz import *


@register_command(
    "monitor action-group show",
)
class Show(AAZCommand):
    """Show the details of an action group.

    :example: Show the details of an action group (commonly used with --output and --query).
        az monitor action-group show --name MyActionGroup --resource-group MyResourceGroup
    """

    _aaz_info = {
        "version": "2024-10-01-preview",
        "resources": [
            ["mgmt-plane", "/subscriptions/{}/resourcegroups/{}/providers/microsoft.insights/actiongroups/{}", "2024-10-01-preview"],
        ]
    }

    def _handler(self, command_args):
        super()._handler(command_args)
        self._execute_operations()
        return self._output()

    _args_schema = None

    @classmethod
    def _build_arguments_schema(cls, *args, **kwargs):
        if cls._args_schema is not None:
            return cls._args_schema
        cls._args_schema = super()._build_arguments_schema(*args, **kwargs)

        # define Arg Group ""

        _args_schema = cls._args_schema
        _args_schema.action_group_name = AAZStrArg(
            options=["-n", "--name", "--action-group-name"],
            help="The name of the action group.",
            required=True,
            id_part="name",
        )
        _args_schema.resource_group = AAZResourceGroupNameArg(
            required=True,
        )
        return cls._args_schema

    def _execute_operations(self):
        self.pre_operations()
        self.ActionGroupsGet(ctx=self.ctx)()
        self.post_operations()

    @register_callback
    def pre_operations(self):
        pass

    @register_callback
    def post_operations(self):
        pass

    def _output(self, *args, **kwargs):
        result = self.deserialize_output(self.ctx.vars.instance, client_flatten=True)
        return result

    class ActionGroupsGet(AAZHttpOperation):
        CLIENT_TYPE = "MgmtClient"

        def __call__(self, *args, **kwargs):
            request = self.make_request()
            session = self.client.send_request(request=request, stream=False, **kwargs)
            if session.http_response.status_code in [200]:
                return self.on_200(session)

            return self.on_error(session.http_response)

        @property
        def url(self):
            return self.client.format_url(
                "/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/Microsoft.Insights/actionGroups/{actionGroupName}",
                **self.url_parameters
            )

        @property
        def method(self):
            return "GET"

        @property
        def error_format(self):
            return "ODataV4Format"

        @property
        def url_parameters(self):
            parameters = {
                **self.serialize_url_param(
                    "actionGroupName", self.ctx.args.action_group_name,
                    required=True,
                ),
                **self.serialize_url_param(
                    "resourceGroupName", self.ctx.args.resource_group,
                    required=True,
                ),
                **self.serialize_url_param(
                    "subscriptionId", self.ctx.subscription_id,
                    required=True,
                ),
            }
            return parameters

        @property
        def query_parameters(self):
            parameters = {
                **self.serialize_query_param(
                    "api-version", "2024-10-01-preview",
                    required=True,
                ),
            }
            return parameters

        @property
        def header_parameters(self):
            parameters = {
                **self.serialize_header_param(
                    "Accept", "application/json",
                ),
            }
            return parameters

        def on_200(self, session):
            data = self.deserialize_http_content(session)
            self.ctx.set_var(
                "instance",
                data,
                schema_builder=self._build_schema_on_200
            )

        _schema_on_200 = None

        @classmethod
        def _build_schema_on_200(cls):
            if cls._schema_on_200 is not None:
                return cls._schema_on_200

            cls._schema_on_200 = AAZObjectType()

            _schema_on_200 = cls._schema_on_200
            _schema_on_200.id = AAZStrType(
                flags={"read_only": True},
            )
            _schema_on_200.identity = AAZIdentityObjectType()
            _schema_on_200.location = AAZStrType(
                flags={"required": True},
            )
            _schema_on_200.name = AAZStrType(
                flags={"read_only": True},
            )
            _schema_on_200.properties = AAZObjectType(
                flags={"client_flatten": True},
            )
            _schema_on_200.tags = AAZDictType()
            _schema_on_200.type = AAZStrType(
                flags={"read_only": True},
            )

            identity = cls._schema_on_200.identity
            identity.principal_id = AAZStrType(
                serialized_name="principalId",
                flags={"read_only": True},
            )
            identity.tenant_id = AAZStrType(
                serialized_name="tenantId",
                flags={"read_only": True},
            )
            identity.type = AAZStrType(
                flags={"required": True},
            )
            identity.user_assigned_identities = AAZDictType(
                serialized_name="userAssignedIdentities",
            )

            user_assigned_identities = cls._schema_on_200.identity.user_assigned_identities
            user_assigned_identities.Element = AAZObjectType(
                nullable=True,
            )

            _element = cls._schema_on_200.identity.user_assigned_identities.Element
            _element.client_id = AAZStrType(
                serialized_name="clientId",
                flags={"read_only": True},
            )
            _element.principal_id = AAZStrType(
                serialized_name="principalId",
                flags={"read_only": True},
            )

            properties = cls._schema_on_200.properties
            properties.arm_role_receivers = AAZListType(
                serialized_name="armRoleReceivers",
            )
            properties.automation_runbook_receivers = AAZListType(
                serialized_name="automationRunbookReceivers",
            )
            properties.azure_app_push_receivers = AAZListType(
                serialized_name="azureAppPushReceivers",
            )
            properties.azure_function_receivers = AAZListType(
                serialized_name="azureFunctionReceivers",
            )
            properties.email_receivers = AAZListType(
                serialized_name="emailReceivers",
            )
            properties.enabled = AAZBoolType(
                flags={"required": True},
            )
            properties.event_hub_receivers = AAZListType(
                serialized_name="eventHubReceivers",
            )
            properties.group_short_name = AAZStrType(
                serialized_name="groupShortName",
                flags={"required": True},
            )
            properties.incident_receivers = AAZListType(
                serialized_name="incidentReceivers",
            )
            properties.itsm_receivers = AAZListType(
                serialized_name="itsmReceivers",
            )
            properties.logic_app_receivers = AAZListType(
                serialized_name="logicAppReceivers",
            )
            properties.sms_receivers = AAZListType(
                serialized_name="smsReceivers",
            )
            properties.voice_receivers = AAZListType(
                serialized_name="voiceReceivers",
            )
            properties.webhook_receivers = AAZListType(
                serialized_name="webhookReceivers",
            )

            arm_role_receivers = cls._schema_on_200.properties.arm_role_receivers
            arm_role_receivers.Element = AAZObjectType()

            _element = cls._schema_on_200.properties.arm_role_receivers.Element
            _element.name = AAZStrType(
                flags={"required": True},
            )
            _element.role_id = AAZStrType(
                serialized_name="roleId",
                flags={"required": True},
            )
            _element.use_common_alert_schema = AAZBoolType(
                serialized_name="useCommonAlertSchema",
            )

            automation_runbook_receivers = cls._schema_on_200.properties.automation_runbook_receivers
            automation_runbook_receivers.Element = AAZObjectType()

            _element = cls._schema_on_200.properties.automation_runbook_receivers.Element
            _element.automation_account_id = AAZStrType(
                serialized_name="automationAccountId",
                flags={"required": True},
            )
            _element.is_global_runbook = AAZBoolType(
                serialized_name="isGlobalRunbook",
                flags={"required": True},
            )
            _element.managed_identity = AAZStrType(
                serialized_name="managedIdentity",
            )
            _element.name = AAZStrType()
            _element.runbook_name = AAZStrType(
                serialized_name="runbookName",
                flags={"required": True},
            )
            _element.service_uri = AAZStrType(
                serialized_name="serviceUri",
            )
            _element.use_common_alert_schema = AAZBoolType(
                serialized_name="useCommonAlertSchema",
            )
            _element.webhook_resource_id = AAZStrType(
                serialized_name="webhookResourceId",
                flags={"required": True},
            )

            azure_app_push_receivers = cls._schema_on_200.properties.azure_app_push_receivers
            azure_app_push_receivers.Element = AAZObjectType()

            _element = cls._schema_on_200.properties.azure_app_push_receivers.Element
            _element.email_address = AAZStrType(
                serialized_name="emailAddress",
                flags={"required": True},
            )
            _element.name = AAZStrType(
                flags={"required": True},
            )

            azure_function_receivers = cls._schema_on_200.properties.azure_function_receivers
            azure_function_receivers.Element = AAZObjectType()

            _element = cls._schema_on_200.properties.azure_function_receivers.Element
            _element.function_app_resource_id = AAZStrType(
                serialized_name="functionAppResourceId",
                flags={"required": True},
            )
            _element.function_name = AAZStrType(
                serialized_name="functionName",
                flags={"required": True},
            )
            _element.http_trigger_url = AAZStrType(
                serialized_name="httpTriggerUrl",
                flags={"required": True},
            )
            _element.managed_identity = AAZStrType(
                serialized_name="managedIdentity",
            )
            _element.name = AAZStrType(
                flags={"required": True},
            )
            _element.use_common_alert_schema = AAZBoolType(
                serialized_name="useCommonAlertSchema",
            )

            email_receivers = cls._schema_on_200.properties.email_receivers
            email_receivers.Element = AAZObjectType()

            _element = cls._schema_on_200.properties.email_receivers.Element
            _element.email_address = AAZStrType(
                serialized_name="emailAddress",
                flags={"required": True},
            )
            _element.name = AAZStrType(
                flags={"required": True},
            )
            _element.status = AAZStrType(
                flags={"read_only": True},
            )
            _element.use_common_alert_schema = AAZBoolType(
                serialized_name="useCommonAlertSchema",
            )

            event_hub_receivers = cls._schema_on_200.properties.event_hub_receivers
            event_hub_receivers.Element = AAZObjectType()

            _element = cls._schema_on_200.properties.event_hub_receivers.Element
            _element.event_hub_name = AAZStrType(
                serialized_name="eventHubName",
                flags={"required": True},
            )
            _element.event_hub_name_space = AAZStrType(
                serialized_name="eventHubNameSpace",
                flags={"required": True},
            )
            _element.managed_identity = AAZStrType(
                serialized_name="managedIdentity",
            )
            _element.name = AAZStrType(
                flags={"required": True},
            )
            _element.subscription_id = AAZStrType(
                serialized_name="subscriptionId",
                flags={"required": True},
            )
            _element.tenant_id = AAZStrType(
                serialized_name="tenantId",
            )
            _element.use_common_alert_schema = AAZBoolType(
                serialized_name="useCommonAlertSchema",
            )

            incident_receivers = cls._schema_on_200.properties.incident_receivers
            incident_receivers.Element = AAZObjectType()

            _element = cls._schema_on_200.properties.incident_receivers.Element
            _element.connection = AAZObjectType(
                flags={"required": True},
            )
            _element.incident_management_service = AAZStrType(
                serialized_name="incidentManagementService",
                flags={"required": True},
            )
            _element.mappings = AAZDictType(
                flags={"required": True},
            )
            _element.name = AAZStrType(
                flags={"required": True},
            )

            connection = cls._schema_on_200.properties.incident_receivers.Element.connection
            connection.id = AAZStrType(
                flags={"required": True},
            )
            connection.name = AAZStrType(
                flags={"required": True},
            )

            mappings = cls._schema_on_200.properties.incident_receivers.Element.mappings
            mappings.Element = AAZStrType()

            itsm_receivers = cls._schema_on_200.properties.itsm_receivers
            itsm_receivers.Element = AAZObjectType()

            _element = cls._schema_on_200.properties.itsm_receivers.Element
            _element.connection_id = AAZStrType(
                serialized_name="connectionId",
                flags={"required": True},
            )
            _element.name = AAZStrType(
                flags={"required": True},
            )
            _element.region = AAZStrType(
                flags={"required": True},
            )
            _element.ticket_configuration = AAZStrType(
                serialized_name="ticketConfiguration",
                flags={"required": True},
            )
            _element.workspace_id = AAZStrType(
                serialized_name="workspaceId",
                flags={"required": True},
            )

            logic_app_receivers = cls._schema_on_200.properties.logic_app_receivers
            logic_app_receivers.Element = AAZObjectType()

            _element = cls._schema_on_200.properties.logic_app_receivers.Element
            _element.callback_url = AAZStrType(
                serialized_name="callbackUrl",
                flags={"required": True},
            )
            _element.managed_identity = AAZStrType(
                serialized_name="managedIdentity",
            )
            _element.name = AAZStrType(
                flags={"required": True},
            )
            _element.resource_id = AAZStrType(
                serialized_name="resourceId",
                flags={"required": True},
            )
            _element.use_common_alert_schema = AAZBoolType(
                serialized_name="useCommonAlertSchema",
            )

            sms_receivers = cls._schema_on_200.properties.sms_receivers
            sms_receivers.Element = AAZObjectType()

            _element = cls._schema_on_200.properties.sms_receivers.Element
            _element.country_code = AAZStrType(
                serialized_name="countryCode",
                flags={"required": True},
            )
            _element.name = AAZStrType(
                flags={"required": True},
            )
            _element.phone_number = AAZStrType(
                serialized_name="phoneNumber",
                flags={"required": True},
            )
            _element.status = AAZStrType(
                flags={"read_only": True},
            )

            voice_receivers = cls._schema_on_200.properties.voice_receivers
            voice_receivers.Element = AAZObjectType()

            _element = cls._schema_on_200.properties.voice_receivers.Element
            _element.country_code = AAZStrType(
                serialized_name="countryCode",
                flags={"required": True},
            )
            _element.name = AAZStrType(
                flags={"required": True},
            )
            _element.phone_number = AAZStrType(
                serialized_name="phoneNumber",
                flags={"required": True},
            )

            webhook_receivers = cls._schema_on_200.properties.webhook_receivers
            webhook_receivers.Element = AAZObjectType()

            _element = cls._schema_on_200.properties.webhook_receivers.Element
            _element.identifier_uri = AAZStrType(
                serialized_name="identifierUri",
            )
            _element.managed_identity = AAZStrType(
                serialized_name="managedIdentity",
            )
            _element.name = AAZStrType(
                flags={"required": True},
            )
            _element.object_id = AAZStrType(
                serialized_name="objectId",
            )
            _element.service_uri = AAZStrType(
                serialized_name="serviceUri",
                flags={"required": True},
            )
            _element.tenant_id = AAZStrType(
                serialized_name="tenantId",
            )
            _element.use_aad_auth = AAZBoolType(
                serialized_name="useAadAuth",
            )
            _element.use_common_alert_schema = AAZBoolType(
                serialized_name="useCommonAlertSchema",
            )

            tags = cls._schema_on_200.tags
            tags.Element = AAZStrType()

            return cls._schema_on_200


class _ShowHelper:
    """Helper class for Show"""


__all__ = ["Show"]



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
#
# Code generated by aaz-dev-tools
# --------------------------------------------------------------------------------------------

# pylint: skip-file
# flake8: noqa

from azure.cli.core.aaz import *


@register_command(
    "policy set-definition update",
)
class Update(AAZCommand):
    """Update a policy set definition.

    Update the policy set definition in the given subscription or management group with the given name by applying the given properties.

    :example: Update a policy set definition
        az policy set-definition update --definitions '[ { 'policyDefinitionId': '/subscriptions/{subscriptionId}/providers/Microsoft.Authorization/policyDefinitions/storagePolicy' } ]' --name MyPolicySetDefinition

    :example: Update the groups and definitions within a policy set definition
        az policy set-definition update -n computeRequirements --definitions "[ { 'policyDefinitionId': '/subscriptions/{subscriptionId}/providers/Microsoft.Authorization/policyDefinitions/storagePolicy', 'groupNames': [ 'CostSaving', 'Organizational' ] }, { 'policyDefinitionId': '/subscriptions/{subscriptionId}/providers/Microsoft.Authorization/policyDefinitions/tagPolicy', 'groupNames': [ 'Organizational' ] } ]" --definition-groups "[{ 'name': 'CostSaving' }, { 'name': 'Organizational' } ]"
    """

    _aaz_info = {
        "version": "2024-05-01",
        "resources": [
            ["mgmt-plane", "/providers/microsoft.management/managementgroups/{}/providers/microsoft.authorization/policysetdefinitions/{}", "2024-05-01"],
            ["mgmt-plane", "/subscriptions/{}/providers/microsoft.authorization/policysetdefinitions/{}", "2024-05-01"],
        ]
    }

    AZ_SUPPORT_GENERIC_UPDATE = True

    def _handler(self, command_args):
        super()._handler(command_args)
        self._execute_operations()
        return self._output()

    _args_schema = None

    @classmethod
    def _build_arguments_schema(cls, *args, **kwargs):
        if cls._args_schema is not None:
            return cls._args_schema
        cls._args_schema = super()._build_arguments_schema(*args, **kwargs)

        # define Arg Group ""

        _args_schema = cls._args_schema
        _args_schema.management_group = AAZStrArg(
            options=["--management-group"],
            help={"short-summary": "The management group.", "long-summary": "The management group with the given name where the policy set definition resides."},
        )
        _args_schema.name = AAZStrArg(
            options=["-n", "--name"],
            help="The name of the policy set definition.",
            required=True,
            fmt=AAZStrArgFormat(
                pattern="^[^<>*%&:\\?.+/]*[^<>*%&:\\?.+/ ]+$",
            ),
        )

        # define Arg Group "Properties"

        _args_schema = cls._args_schema
        _args_schema.description = AAZStrArg(
            options=["--description"],
            arg_group="Properties",
            help={"short-summary": "Policy set definition description.", "long-summary": "Full description of the policy set definition."},
            nullable=True,
        )
        _args_schema.display_name = AAZStrArg(
            options=["--display-name"],
            arg_group="Properties",
            help={"short-summary": "The display name of the policy set definition.", "long-summary": "The display name of the policy set definition is not part of its ID, allowing for longer and more flexible naming."},
            nullable=True,
        )
        _args_schema.metadata = AAZDictArg(
            options=["--metadata"],
            arg_group="Properties",
            help={"short-summary": "The policy set definition metadata.", "long-summary": "The policy set definition metadata. Metadata is an open-ended object and is typically a collection of key value pairs."},
            nullable=True,
        )
        _args_schema.params = AAZDictArg(
            options=["-p", "--params"],
            arg_group="Properties",
            help={"short-summary": "The policy set definition parameter definitions.", "long-summary": "The definitions for parameters used in the policy rule. The keys are the parameter names."},
            nullable=True,
        )
        _args_schema.definition_groups = AAZListArg(
            options=["--definition-groups"],
            arg_group="Properties",
            help="The metadata describing groups of policy definition references within the policy set definition.",
            nullable=True,
        )
        _args_schema.definitions = AAZListArg(
            options=["--definitions"],
            arg_group="Properties",
            help="An array of policy definition references.",
        )
        _args_schema.version = AAZStrArg(
            options=["--version"],
            arg_group="Properties",
            help={"short-summary": "The policy set definition version.", "long-summary": "The policy set definition version in #.#.# format."},
            nullable=True,
        )

        metadata = cls._args_schema.metadata
        metadata.Element = AAZAnyTypeArg(
            nullable=True,
        )

        params = cls._args_schema.params
        params.Element = AAZObjectArg(
            nullable=True,
        )

        _element = cls._args_schema.params.Element
        _element.allowed_values = AAZListArg(
            options=["allowed-values"],
            help="The allowed values for the parameter.",
            nullable=True,
        )
        _element.default_value = AAZAnyTypeArg(
            options=["default-value"],
            help="The default value for the parameter if no value is provided.",
            nullable=True,
        )
        _element.metadata = AAZFreeFormDictArg(
            options=["metadata"],
            help="General metadata for the parameter.",
            nullable=True,
        )
        _element.schema = AAZDictArg(
            options=["schema"],
            help="Provides validation of parameter inputs during assignment using a self-defined JSON schema. This property is only supported for object-type parameters and follows the Json.NET Schema 2019-09 implementation. You can learn more about using schemas at https://json-schema.org/ and test draft schemas at https://www.jsonschemavalidator.net/.",
            nullable=True,
        )
        _element.type = AAZStrArg(
            options=["type"],
            help="The data type of the parameter.",
            nullable=True,
            enum={"Array": "Array", "Boolean": "Boolean", "DateTime": "DateTime", "Float": "Float", "Integer": "Integer", "Object": "Object", "String": "String"},
        )

        allowed_values = cls._args_schema.params.Element.allowed_values
        allowed_values.Element = AAZAnyTypeArg(
            nullable=True,
        )

        schema = cls._args_schema.params.Element.schema
        schema.Element = AAZAnyTypeArg(
            nullable=True,
        )

        definition_groups = cls._args_schema.definition_groups
        definition_groups.Element = AAZObjectArg(
            nullable=True,
        )

        _element = cls._args_schema.definition_groups.Element
        _element.additional_metadata_id = AAZStrArg(
            options=["additional-metadata-id"],
            help="A resource ID of a resource that contains additional metadata about the group.",
            nullable=True,
        )
        _element.category = AAZStrArg(
            options=["category"],
            help="The group's category.",
            nullable=True,
        )
        _element.description = AAZStrArg(
            options=["description"],
            help="The group's description.",
            nullable=True,
        )
        _element.display_name = AAZStrArg(
            options=["display-name"],
            help="The group's display name.",
            nullable=True,
        )
        _element.name = AAZStrArg(
            options=["name"],
            help="The name of the group.",
        )

        definitions = cls._args_schema.definitions
        definitions.Element = AAZObjectArg(
            nullable=True,
        )

        _element = cls._args_schema.definitions.Element
        _element.definition_version = AAZStrArg(
            options=["definition-version"],
            help="The version of the policy definition to use.",
            nullable=True,
        )
        _element.group_names = AAZListArg(
            options=["group-names"],
            help="The name of the groups that this policy definition reference belongs to.",
            nullable=True,
        )
        _element.parameters = AAZDictArg(
            options=["parameters"],
            help="The parameter values for the referenced policy rule. The keys are the parameter names.",
            nullable=True,
        )
        _element.policy_definition_id = AAZStrArg(
            options=["policy-definition-id"],
            help="The ID of the policy definition or policy set definition.",
        )
        _element.policy_definition_reference_id = AAZStrArg(
            options=["policy-definition-reference-id"],
            help="A unique id (within the policy set definition) for this policy definition reference.",
            nullable=True,
        )

        group_names = cls._args_schema.definitions.Element.group_names
        group_names.Element = AAZStrArg(
            nullable=True,
        )

        parameters = cls._args_schema.definitions.Element.parameters
        parameters.Element = AAZObjectArg(
            nullable=True,
        )

        _element = cls._args_schema.definitions.Element.parameters.Element
        _element.value = AAZAnyTypeArg(
            options=["value"],
            help="The value of the parameter.",
            nullable=True,
        )
        return cls._args_schema

    def _execute_operations(self):
        self.pre_operations()
        condition_0 = has_value(self.ctx.args.management_group) and has_value(self.ctx.args.name)
        condition_1 = has_value(self.ctx.args.name) and has_value(self.ctx.subscription_id)
        condition_2 = has_value(self.ctx.args.management_group) and has_value(self.ctx.args.name)
        condition_3 = has_value(self.ctx.args.name) and has_value(self.ctx.subscription_id)
        if condition_0:
            self.PolicySetDefinitionsGetAtManagementGroup(ctx=self.ctx)()
        if condition_1:
            self.PolicySetDefinitionsGet(ctx=self.ctx)()
        self.pre_instance_update(self.ctx.vars.instance)
        self.InstanceUpdateByJson(ctx=self.ctx)()
        self.InstanceUpdateByGeneric(ctx=self.ctx)()
        self.post_instance_update(self.ctx.vars.instance)
        if condition_2:
            self.PolicySetDefinitionsCreateOrUpdateAtManagementGroup(ctx=self.ctx)()
        if condition_3:
            self.PolicySetDefinitionsCreateOrUpdate(ctx=self.ctx)()
        self.post_operations()

    @register_callback
    def pre_operations(self):
        pass

    @register_callback
    def post_operations(self):
        pass

    @register_callback
    def pre_instance_update(self, instance):
        pass

    @register_callback
    def post_instance_update(self, instance):
        pass

    def _output(self, *args, **kwargs):
        result = self.deserialize_output(self.ctx.vars.instance, client_flatten=True)
        return result

    class PolicySetDefinitionsGetAtManagementGroup(AAZHttpOperation):
        CLIENT_TYPE = "MgmtClient"

        def __call__(self, *args, **kwargs):
            request = self.make_request()
            session = self.client.send_request(request=request, stream=False, **kwargs)
            if session.http_response.status_code in [200]:
                return self.on_200(session)

            return self.on_error(session.http_response)

        @property
        def url(self):
            return self.client.format_url(
                "/providers/Microsoft.Management/managementGroups/{managementGroupId}/providers/Microsoft.Authorization/policySetDefinitions/{policySetDefinitionName}",
                **self.url_parameters
            )

        @property
        def method(self):
            return "GET"

        @property
        def error_format(self):
            return "MgmtErrorFormat"

        @property
        def url_parameters(self):
            parameters = {
                **self.serialize_url_param(
                    "managementGroupId", self.ctx.args.management_group,
                    required=True,
                ),
                **self.serialize_url_param(
                    "policySetDefinitionName", self.ctx.args.name,
                    required=True,
                ),
            }
            return parameters

        @property
        def query_parameters(self):
            parameters = {
                **self.serialize_query_param(
                    "api-version", "2024-05-01",
                    required=True,
                ),
            }
            return parameters

        @property
        def header_parameters(self):
            parameters = {
                **self.serialize_header_param(
                    "Accept", "application/json",
                ),
            }
            return parameters

        def on_200(self, session):
            data = self.deserialize_http_content(session)
            self.ctx.set_var(
                "instance",
                data,
                schema_builder=self._build_schema_on_200
            )

        _schema_on_200 = None

        @classmethod
        def _build_schema_on_200(cls):
            if cls._schema_on_200 is not None:
                return cls._schema_on_200

            cls._schema_on_200 = AAZObjectType()
            _UpdateHelper._build_schema_policy_set_definition_read(cls._schema_on_200)

            return cls._schema_on_200

    class PolicySetDefinitionsGet(AAZHttpOperation):
        CLIENT_TYPE = "MgmtClient"

        def __call__(self, *args, **kwargs):
            request = self.make_request()
            session = self.client.send_request(request=request, stream=False, **kwargs)
            if session.http_response.status_code in [200]:
                return self.on_200(session)

            return self.on_error(session.http_response)

        @property
        def url(self):
            return self.client.format_url(
                "/subscriptions/{subscriptionId}/providers/Microsoft.Authorization/policySetDefinitions/{policySetDefinitionName}",
                **self.url_parameters
            )

        @property
        def method(self):
            return "GET"

        @property
        def error_format(self):
            return "MgmtErrorFormat"

        @property
        def url_parameters(self):
            parameters = {
                **self.serialize_url_param(
                    "policySetDefinitionName", self.ctx.args.name,
                    required=True,
                ),
                **self.serialize_url_param(
                    "subscriptionId", self.ctx.subscription_id,
                    required=True,
                ),
            }
            return parameters

        @property
        def query_parameters(self):
            parameters = {
                **self.serialize_query_param(
                    "api-version", "2024-05-01",
                    required=True,
                ),
            }
            return parameters

        @property
        def header_parameters(self):
            parameters = {
                **self.serialize_header_param(
                    "Accept", "application/json",
                ),
            }
            return parameters

        def on_200(self, session):
            data = self.deserialize_http_content(session)
            self.ctx.set_var(
                "instance",
                data,
                schema_builder=self._build_schema_on_200
            )

        _schema_on_200 = None

        @classmethod
        def _build_schema_on_200(cls):
            if cls._schema_on_200 is not None:
                return cls._schema_on_200

            cls._schema_on_200 = AAZObjectType()
            _UpdateHelper._build_schema_policy_set_definition_read(cls._schema_on_200)

            return cls._schema_on_200

    class PolicySetDefinitionsCreateOrUpdateAtManagementGroup(AAZHttpOperation):
        CLIENT_TYPE = "MgmtClient"

        def __call__(self, *args, **kwargs):
            request = self.make_request()
            session = self.client.send_request(request=request, stream=False, **kwargs)
            if session.http_response.status_code in [200, 201]:
                return self.on_200_201(session)

            return self.on_error(session.http_response)

        @property
        def url(self):
            return self.client.format_url(
                "/providers/Microsoft.Management/managementGroups/{managementGroupId}/providers/Microsoft.Authorization/policySetDefinitions/{policySetDefinitionName}",
                **self.url_parameters
            )

        @property
        def method(self):
            return "PUT"

        @property
        def error_format(self):
            return "MgmtErrorFormat"

        @property
        def url_parameters(self):
            parameters = {
                **self.serialize_url_param(
                    "managementGroupId", self.ctx.args.management_group,
                    required=True,
                ),
                **self.serialize_url_param(
                    "policySetDefinitionName", self.ctx.args.name,
                    required=True,
                ),
            }
            return parameters

        @property
        def query_parameters(self):
            parameters = {
                **self.serialize_query_param(
                    "api-version", "2024-05-01",
                    required=True,
                ),
            }
            return parameters

        @property
        def header_parameters(self):
            parameters = {
                **self.serialize_header_param(
                    "Content-Type", "application/json",
                ),
                **self.serialize_header_param(
                    "Accept", "application/json",
                ),
            }
            return parameters

        @property
        def content(self):
            _content_value, _builder = self.new_content_builder(
                self.ctx.args,
                value=self.ctx.vars.instance,
            )

            return self.serialize_content(_content_value)

        def on_200_201(self, session):
            data = self.deserialize_http_content(session)
            self.ctx.set_var(
                "instance",
                data,
                schema_builder=self._build_schema_on_200_201
            )

        _schema_on_200_201 = None

        @classmethod
        def _build_schema_on_200_201(cls):
            if cls._schema_on_200_201 is not None:
                return cls._schema_on_200_201

            cls._schema_on_200_201 = AAZObjectType()
            _UpdateHelper._build_schema_policy_set_definition_read(cls._schema_on_200_201)

            return cls._schema_on_200_201

    class PolicySetDefinitionsCreateOrUpdate(AAZHttpOperation):
        CLIENT_TYPE = "MgmtClient"

        def __call__(self, *args, **kwargs):
            request = self.make_request()
            session = self.client.send_request(request=request, stream=False, **kwargs)
            if session.http_response.status_code in [200, 201]:
                return self.on_200_201(session)

            return self.on_error(session.http_response)

        @property
        def url(self):
            return self.client.format_url(
                "/subscriptions/{subscriptionId}/providers/Microsoft.Authorization/policySetDefinitions/{policySetDefinitionName}",
                **self.url_parameters
            )

        @property
        def method(self):
            return "PUT"

        @property
        def error_format(self):
            return "MgmtErrorFormat"

        @property
        def url_parameters(self):
            parameters = {
                **self.serialize_url_param(
                    "policySetDefinitionName", self.ctx.args.name,
                    required=True,
                ),
                **self.serialize_url_param(
                    "subscriptionId", self.ctx.subscription_id,
                    required=True,
                ),
            }
            return parameters

        @property
        def query_parameters(self):
            parameters = {
                **self.serialize_query_param(
                    "api-version", "2024-05-01",
                    required=True,
                ),
            }
            return parameters

        @property
        def header_parameters(self):
            parameters = {
                **self.serialize_header_param(
                    "Content-Type", "application/json",
                ),
                **self.serialize_header_param(
                    "Accept", "application/json",
                ),
            }
            return parameters

        @property
        def content(self):
            _content_value, _builder = self.new_content_builder(
                self.ctx.args,
                value=self.ctx.vars.instance,
            )

            return self.serialize_content(_content_value)

        def on_200_201(self, session):
            data = self.deserialize_http_content(session)
            self.ctx.set_var(
                "instance",
                data,
                schema_builder=self._build_schema_on_200_201
            )

        _schema_on_200_201 = None

        @classmethod
        def _build_schema_on_200_201(cls):
            if cls._schema_on_200_201 is not None:
                return cls._schema_on_200_201

            cls._schema_on_200_201 = AAZObjectType()
            _UpdateHelper._build_schema_policy_set_definition_read(cls._schema_on_200_201)

            return cls._schema_on_200_201

    class InstanceUpdateByJson(AAZJsonInstanceUpdateOperation):

        def __call__(self, *args, **kwargs):
            self._update_instance(self.ctx.vars.instance)

        def _update_instance(self, instance):
            _instance_value, _builder = self.new_content_builder(
                self.ctx.args,
                value=instance,
                typ=AAZObjectType
            )
            _builder.set_prop("properties", AAZObjectType, typ_kwargs={"flags": {"client_flatten": True}})

            properties = _builder.get(".properties")
            if properties is not None:
                properties.set_prop("description", AAZStrType, ".description")
                properties.set_prop("displayName", AAZStrType, ".display_name")
                properties.set_prop("metadata", AAZDictType, ".metadata")
                properties.set_prop("parameters", AAZDictType, ".params")
                properties.set_prop("policyDefinitionGroups", AAZListType, ".definition_groups")
                properties.set_prop("policyDefinitions", AAZListType, ".definitions", typ_kwargs={"flags": {"required": True}})
                properties.set_prop("version", AAZStrType, ".version")

            metadata = _builder.get(".properties.metadata")
            if metadata is not None:
                metadata.set_elements(AAZAnyType, ".")

            parameters = _builder.get(".properties.parameters")
            if parameters is not None:
                parameters.set_elements(AAZObjectType, ".")

            _elements = _builder.get(".properties.parameters{}")
            if _elements is not None:
                _elements.set_prop("allowedValues", AAZListType, ".allowed_values")
                _elements.set_prop("defaultValue", AAZAnyType, ".default_value")
                _elements.set_prop("metadata", AAZFreeFormDictType, ".metadata")
                _elements.set_prop("schema", AAZDictType, ".schema")
                _elements.set_prop("type", AAZStrType, ".type")

            allowed_values = _builder.get(".properties.parameters{}.allowedValues")
            if allowed_values is not None:
                allowed_values.set_elements(AAZAnyType, ".")

            metadata = _builder.get(".properties.parameters{}.metadata")
            if metadata is not None:
                metadata.set_anytype_elements(".")

            schema = _builder.get(".properties.parameters{}.schema")
            if schema is not None:
                schema.set_elements(AAZAnyType, ".")

            policy_definition_groups = _builder.get(".properties.policyDefinitionGroups")
            if policy_definition_groups is not None:
                policy_definition_groups.set_elements(AAZObjectType, ".")

            _elements = _builder.get(".properties.policyDefinitionGroups[]")
            if _elements is not None:
                _elements.set_prop("additionalMetadataId", AAZStrType, ".additional_metadata_id")
                _elements.set_prop("category", AAZStrType, ".category")
                _elements.set_prop("description", AAZStrType, ".description")
                _elements.set_prop("displayName", AAZStrType, ".display_name")
                _elements.set_prop("name", AAZStrType, ".name", typ_kwargs={"flags": {"required": True}})

            policy_definitions = _builder.get(".properties.policyDefinitions")
            if policy_definitions is not None:
                policy_definitions.set_elements(AAZObjectType, ".")

            _elements = _builder.get(".properties.policyDefinitions[]")
            if _elements is not None:
                _elements.set_prop("definitionVersion", AAZStrType, ".definition_version")
                _elements.set_prop("groupNames", AAZListType, ".group_names")
                _elements.set_prop("parameters", AAZDictType, ".parameters")
                _elements.set_prop("policyDefinitionId", AAZStrType, ".policy_definition_id", typ_kwargs={"flags": {"required": True}})
                _elements.set_prop("policyDefinitionReferenceId", AAZStrType, ".policy_definition_reference_id")

            group_names = _builder.get(".properties.policyDefinitions[].groupNames")
            if group_names is not None:
                group_names.set_elements(AAZStrType, ".")

            parameters = _builder.get(".properties.policyDefinitions[].parameters")
            if parameters is not None:
                parameters.set_elements(AAZObjectType, ".")

            _elements = _builder.get(".properties.policyDefinitions[].parameters{}")
            if _elements is not None:
                _elements.set_prop("value", AAZAnyType, ".value")

            return _instance_value

    class InstanceUpdateByGeneric(AAZGenericInstanceUpdateOperation):

        def __call__(self, *args, **kwargs):
            self._update_instance_by_generic(
                self.ctx.vars.instance,
                self.ctx.generic_update_args
            )


class _UpdateHelper:
    """Helper class for Update"""

    _schema_policy_set_definition_read = None

    @classmethod
    def _build_schema_policy_set_definition_read(cls, _schema):
        if cls._schema_policy_set_definition_read is not None:
            _schema.id = cls._schema_policy_set_definition_read.id
            _schema.name = cls._schema_policy_set_definition_read.name
            _schema.properties = cls._schema_policy_set_definition_read.properties
            _schema.system_data = cls._schema_policy_set_definition_read.system_data
            _schema.type = cls._schema_policy_set_definition_read.type
            return

        cls._schema_policy_set_definition_read = _schema_policy_set_definition_read = AAZObjectType()

        policy_set_definition_read = _schema_policy_set_definition_read
        policy_set_definition_read.id = AAZStrType(
            flags={"read_only": True},
        )
        policy_set_definition_read.name = AAZStrType(
            flags={"read_only": True},
        )
        policy_set_definition_read.properties = AAZObjectType(
            flags={"client_flatten": True},
        )
        policy_set_definition_read.system_data = AAZObjectType(
            serialized_name="systemData",
            flags={"read_only": True},
        )
        policy_set_definition_read.type = AAZStrType(
            flags={"read_only": True},
        )

        properties = _schema_policy_set_definition_read.properties
        properties.description = AAZStrType()
        properties.display_name = AAZStrType(
            serialized_name="displayName",
        )
        properties.metadata = AAZDictType()
        properties.parameters = AAZDictType()
        properties.policy_definition_groups = AAZListType(
            serialized_name="policyDefinitionGroups",
        )
        properties.policy_definitions = AAZListType(
            serialized_name="policyDefinitions",
            flags={"required": True},
        )
        properties.policy_type = AAZStrType(
            serialized_name="policyType",
        )
        properties.version = AAZStrType()
        properties.versions = AAZListType()

        metadata = _schema_policy_set_definition_read.properties.metadata
        metadata.Element = AAZAnyType()

        parameters = _schema_policy_set_definition_read.properties.parameters
        parameters.Element = AAZObjectType()

        _element = _schema_policy_set_definition_read.properties.parameters.Element
        _element.allowed_values = AAZListType(
            serialized_name="allowedValues",
        )
        _element.default_value = AAZAnyType(
            serialized_name="defaultValue",
        )
        _element.metadata = AAZFreeFormDictType()
        _element.schema = AAZDictType()
        _element.type = AAZStrType()

        allowed_values = _schema_policy_set_definition_read.properties.parameters.Element.allowed_values
        allowed_values.Element = AAZAnyType()

        schema = _schema_policy_set_definition_read.properties.parameters.Element.schema
        schema.Element = AAZAnyType()

        policy_definition_groups = _schema_policy_set_definition_read.properties.policy_definition_groups
        policy_definition_groups.Element = AAZObjectType()

        _element = _schema_policy_set_definition_read.properties.policy_definition_groups.Element
        _element.additional_metadata_id = AAZStrType(
            serialized_name="additionalMetadataId",
        )
        _element.category = AAZStrType()
        _element.description = AAZStrType()
        _element.display_name = AAZStrType(
            serialized_name="displayName",
        )
        _element.name = AAZStrType(
            flags={"required": True},
        )

        policy_definitions = _schema_policy_set_definition_read.properties.policy_definitions
        policy_definitions.Element = AAZObjectType()

        _element = _schema_policy_set_definition_read.properties.policy_definitions.Element
        _element.definition_version = AAZStrType(
            serialized_name="definitionVersion",
        )
        _element.effective_definition_version = AAZStrType(
            serialized_name="effectiveDefinitionVersion",
            flags={"read_only": True},
        )
        _element.group_names = AAZListType(
            serialized_name="groupNames",
        )
        _element.latest_definition_version = AAZStrType(
            serialized_name="latestDefinitionVersion",
            flags={"read_only": True},
        )
        _element.parameters = AAZDictType()
        _element.policy_definition_id = AAZStrType(
            serialized_name="policyDefinitionId",
            flags={"required": True},
        )
        _element.policy_definition_reference_id = AAZStrType(
            serialized_name="policyDefinitionReferenceId",
        )

        group_names = _schema_policy_set_definition_read.properties.policy_definitions.Element.group_names
        group_names.Element = AAZStrType()

        parameters = _schema_policy_set_definition_read.properties.policy_definitions.Element.parameters
        parameters.Element = AAZObjectType()

        _element = _schema_policy_set_definition_read.properties.policy_definitions.Element.parameters.Element
        _element.value = AAZAnyType()

        versions = _schema_policy_set_definition_read.properties.versions
        versions.Element = AAZStrType()

        system_data = _schema_policy_set_definition_read.system_data
        system_data.created_at = AAZStrType(
            serialized_name="createdAt",
        )
        system_data.created_by = AAZStrType(
            serialized_name="createdBy",
        )
        system_data.created_by_type = AAZStrType(
            serialized_name="createdByType",
        )
        system_data.last_modified_at = AAZStrType(
            serialized_name="lastModifiedAt",
        )
        system_data.last_modified_by = AAZStrType(
            serialized_name="lastModifiedBy",
        )
        system_data.last_modified_by_type = AAZStrType(
            serialized_name="lastModifiedByType",
        )

        _schema.id = cls._schema_policy_set_definition_read.id
        _schema.name = cls._schema_policy_set_definition_read.name
        _schema.properties = cls._schema_policy_set_definition_read.properties
        _schema.system_data = cls._schema_policy_set_definition_read.system_data
        _schema.type = cls._schema_policy_set_definition_read.type


__all__ = ["Update"]



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from azure.cli.core.commands.transform import build_table_output
from azure.cli.core.azclierror import CLIInternalError
from knack.util import todict


def transform_support_types(result):
    for res in result:
        res['auth_types'] = ', '.join(res['auth_types'])
        res['client_types'] = ', '.join(res['client_types'])

    return build_table_output(result, [
        ('Source', 'source'),
        ('Target', 'target'),
        ('AuthType', 'auth_types'),
        ('ClientType', 'client_types'),
    ])


def transform_linkers_properties(result):
    from azure.core.polling import LROPoller
    from ._utils import (
        is_aks_linker_by_id,
        get_aks_resource_name
    )

    if isinstance(result, LROPoller):
        result = result.result()

    linkers = [todict(res) for res in result]
    for linker in linkers:
        resource_id = linker.get('id')
        if is_aks_linker_by_id(resource_id):
            linker['kubernetesResourceName'] = get_aks_resource_name(linker)

    return linkers


def transform_linker_properties(result):
    from azure.core.polling import LROPoller
    from ._utils import (
        run_cli_cmd,
        is_aks_linker_by_id,
        get_aks_resource_name
    )

    if result is None:
        return result

    # manually polling if result is a poller
    if isinstance(result, LROPoller):
        result = result.result()

    result = todict(result)
    resource_id = result.get('id')
    if is_aks_linker_by_id(resource_id):
        result['kubernetesResourceName'] = get_aks_resource_name(result)
    try:
        output = run_cli_cmd('az webapp connection list-configuration --id "{}" -o json'.format(resource_id))
        result['configurations'] = output.get('configurations')
    except CLIInternalError:
        pass
    return result


def transform_local_linker_properties(result):
    from azure.core.polling import LROPoller
    from ._utils import (
        run_cli_cmd
    )
    if result is None:
        return result
    # manually polling if result is a poller
    if isinstance(result, LROPoller):
        result = result.result()

    result = todict(result)
    resource_id = result.get('id')
    try:
        output = run_cli_cmd('az connection generate-configuration --id "{}" -o json'.format(resource_id))
        result['configurations'] = output.get('configurations')
    except CLIInternalError:
        pass
    return result


def transform_validation_result(result):
    from azure.core.polling import LROPoller

    # manually polling if result is a poller
    if isinstance(result, LROPoller):
        result = result.result()

    result = todict(result)
    try:
        return result['validationDetail'] or result['additionalProperties']['properties']['validationDetail']
    except Exception:  # pylint: disable=broad-except
        return result



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from azure.cli.core.commands.parameters import (
    get_enum_type,
    get_location_type,
    get_three_state_flag
)
from azure.cli.core.commands.validators import get_default_location_from_resource_group

from ._validators import (
    validate_connstr_props,
    validate_params,
    validate_kafka_params,
    validate_local_params,
    get_default_object_id_of_current_user
)
from ._resource_config import (
    AUTH_TYPE,
    RESOURCE,
    SOURCE_RESOURCES_PARAMS,
    SOURCE_RESOURCES_CREATE_PARAMS,
    TARGET_RESOURCES_PARAMS,
    SOURCE_RESOURCES_OPTIONAL_PARAMS,
    TARGET_RESOURCES_CONNECTION_STRING,
    AUTH_TYPE_PARAMS,
    SUPPORTED_AUTH_TYPE,
    SUPPORTED_CLIENT_TYPE,
    TARGET_SUPPORT_SERVICE_ENDPOINT,
    TARGET_SUPPORT_PRIVATE_ENDPOINT,
    LOCAL_CONNECTION_PARAMS,
    OPT_OUT_OPTION
)
from ._addon_factory import AddonFactory
from knack.arguments import CLIArgumentType
from .action import AddAdditionalConnectionStringProperties, AddCustomizedKeys


def add_source_resource_block(context, source, enable_id=True, target=None):
    source_args = SOURCE_RESOURCES_PARAMS.get(source)
    for resource, args in SOURCE_RESOURCES_PARAMS.items():
        if resource != source:
            for arg in args:
                if arg not in source_args:
                    context.ignore(arg)

    required_args = []
    for arg, content in SOURCE_RESOURCES_PARAMS.get(source).items():
        id_arg = '\'--id\'' if enable_id else '\'--source-id\''
        deprecate_info = content.get('deprecate_info')
        context.argument(arg, configured_default=content.get('configured_default'),
                         options_list=content.get('options'), type=str,
                         deprecate_info=context.deprecate() if deprecate_info else None,
                         help='{}. Required if {} is not specified.{}'.format(
                             content.get('help'), id_arg, deprecate_info))
        required_args.append(content.get('options')[0])

    validator_kwargs = {
        'validator': validate_params}
    if not enable_id:
        context.argument('source_id', options_list=['--source-id'], type=str,
                         help="The resource id of a {source}. Required if {required_args} "
                         "are not specified.".format(
            source=source.value, required_args=str(required_args)),
            **validator_kwargs)
    else:
        required_args.append('--connection')
        context.argument('indentifier', options_list=['--id'], type=str,
                         help="The resource id of the connection. {required_args} are required "
                         "if '--id' is not specified.".format(required_args=str(required_args)))
        context.ignore('source_id')

    # scope parameter
    if source == RESOURCE.KubernetesCluster:
        context.argument('scope', options_list=['--kube-namespace'], type=str, default='default',
                         help="The kubernetes namespace where the connection information "
                         "will be saved into (as kubernetes secret).")
        context.argument('enable_csi', options_list=['--enable-csi'], arg_type=get_three_state_flag(),
                         help="Use keyvault as a secrets store via a CSI volume. "
                         "If specified, AuthType Arguments are not needed.")
        if target == RESOURCE.AppConfig:
            context.argument('enable_appconfig_extension', options_list=['--use-appconfig-extension', '-e'],
                             arg_type=get_three_state_flag(),
                             help="Install Azure App Configuration extension in the Kubernetes cluster.")
        else:
            context.ignore('enable_appconfig_extension')
    elif source == RESOURCE.ContainerApp:
        for arg, content in SOURCE_RESOURCES_CREATE_PARAMS.get(source).items():
            context.argument(arg, options_list=content.get(
                'options'), type=str, help=content.get('help'))
        context.ignore('enable_csi')
        context.ignore('enable_appconfig_extension')
    else:
        context.ignore('scope')
        context.ignore('enable_csi')
        context.ignore('enable_appconfig_extension')

    # slot parameter
    for resource, args in SOURCE_RESOURCES_OPTIONAL_PARAMS.items():
        for arg in args:
            if resource == source:
                deprecate_info = args.get(arg).get('deprecate_info')
                context.argument(arg, options_list=args.get(arg).get(
                    'options'), type=str,
                    deprecate_info=context.deprecate() if deprecate_info else None,
                    help=args.get(arg).get('help'))
            else:
                context.ignore(arg)


def add_auth_block(context, source, target):
    support_auth_types = SUPPORTED_AUTH_TYPE.get(
        source, {}).get(target, [])
    for auth_type in AUTH_TYPE_PARAMS:
        if auth_type in support_auth_types:
            validator = None
            if auth_type == AUTH_TYPE.UserAccount:
                validator = get_default_object_id_of_current_user
            for arg, params in AUTH_TYPE_PARAMS.get(auth_type).items():
                context.argument(arg, options_list=params.get('options'), action=params.get('action'), nargs='*',
                                 help=params.get('help'), arg_group='AuthType', validator=validator)
        else:
            for arg in AUTH_TYPE_PARAMS.get(auth_type):
                context.ignore(arg)


def add_local_connection_block(context, show_id=True):
    context.argument('location', arg_type=CLIArgumentType(
        arg_type=get_location_type(context),
        required=False,
        validator=get_default_location_from_resource_group))

    if show_id:
        context.argument('id', options_list=[
            '--id'], type=str, help='The id of connection.', validator=validate_local_params)
        params = LOCAL_CONNECTION_PARAMS.get('connection_name')
        context.argument('connection_name', options_list=params.get('options'), type=params.get('type'),
                         help=params.get('help'), validator=validate_local_params)


def add_target_resource_block(context, target):
    target_args = TARGET_RESOURCES_PARAMS.get(target)
    for resource, args in TARGET_RESOURCES_PARAMS.items():
        if resource != target:
            for arg in args:
                if arg not in target_args:
                    context.ignore(arg)

    required_args = []
    if target in TARGET_RESOURCES_PARAMS:
        for arg, content in TARGET_RESOURCES_PARAMS.get(target).items():
            context.argument(arg, options_list=content.get('options'), type=str,
                             help='{}. Required if \'--target-id\' is not specified.'.format(content.get('help')))
            required_args.append(content.get('options')[0])
        if target == RESOURCE.NeonPostgres or target == RESOURCE.MongoDbAtlas:
            context.ignore('target_id')
        else:
            context.argument('target_id', type=str,
                             help='The resource id of target service. Required if {required_args} '
                             'are not specified.'.format(required_args=str(required_args)))

    if target != RESOURCE.KeyVault:
        context.ignore('enable_csi')


def add_connection_name_argument(context, source):
    context.argument('connection_name', options_list=['--connection'], type=str,
                     help='Name of the {} connection.'.format(source.value), validator=validate_params)


def add_client_type_argument(context, source, target):
    client_types = SUPPORTED_CLIENT_TYPE.get(source).get(target, [])
    client_types = [item.value for item in client_types]
    context.argument('client_type', options_list=['--client-type'], arg_type=get_enum_type(client_types),
                     help='The client type used on the {}'.format(source.value))


def add_customized_keys_argument(context):
    context.argument('customized_keys', options_list=['--customized-keys'], action=AddCustomizedKeys, nargs='*',
                     help='The customized keys used to change default configuration names. Key is the original '
                     'name, value is the customized name.')


def add_connstr_props_argument(context):
    # linter: length '--additional-connection-string-properties' longer than 22, so use abbreviation
    context.argument('connstr_props', options_list=['--connstr-props'],
                     action=AddAdditionalConnectionStringProperties, nargs='*',
                     help='The additional connection string properties used to build connection string.',
                     validator=validate_connstr_props)


def add_target_type_argument(context, source):
    TARGET_TYPES = [
        elem.value for elem in SUPPORTED_AUTH_TYPE.get(source).keys()]
    context.argument('target_resource_type', options_list=['--target-type', '-t'],
                     arg_type=get_enum_type(TARGET_TYPES), help='The target resource type')


def add_new_addon_argument(context, source, target):
    if AddonFactory.get(target, None):
        context.argument('new_addon', options_list=['--new'], arg_type=get_three_state_flag(), default=False,
                         help='Indicates whether to create a new {} when '
                         'creating the {} connection'.format(target.value, source.value))
    else:
        context.ignore('new_addon')


def add_secret_store_argument(context, source=""):
    if source == RESOURCE.KubernetesCluster:
        context.ignore('key_vault_id')
    else:
        context.argument('key_vault_id', options_list=[
            '--vault-id'], help='The id of key vault to store secret value')


def add_configuration_store_argument(context):
    context.argument('app_config_id', options_list=[
        '--appconfig-id'], help='The app configuration id to store configuration')


def add_vnet_block(context, target):
    if target not in TARGET_SUPPORT_SERVICE_ENDPOINT:
        context.ignore('service_endpoint')
    else:
        context.argument('service_endpoint', options_list=['--service-endpoint'], arg_type=get_three_state_flag(),
                         default=None, arg_group='NetworkSolution',
                         help='Connect target service by service endpoint. Source resource must be in the VNet'
                         ' and target SKU must support service endpoint feature.')

    if target not in TARGET_SUPPORT_PRIVATE_ENDPOINT:
        context.ignore('private_endpoint')
    else:
        context.argument('private_endpoint', options_list=['--private-endpoint'], arg_type=get_three_state_flag(),
                         default=None, arg_group='NetworkSolution',
                         help='Connect target service by private endpoint. '
                         'The private endpoint in source virtual network must be created ahead.')


def add_connection_string_argument(context, source, target):
    if source == RESOURCE.WebApp and target in TARGET_RESOURCES_CONNECTION_STRING:
        context.argument('store_in_connection_string', options_list=['--config-connstr'],
                         arg_type=get_three_state_flag(), default=False, is_preview=True,
                         help='Store configuration into connection strings, '
                         'only could be used together with dotnet client_type')
    else:
        context.ignore('store_in_connection_string')


def add_confluent_kafka_argument(context):
    context.argument('bootstrap_server', options_list=[
        '--bootstrap-server'], help='Kafka bootstrap server url')
    context.argument('kafka_key', options_list=[
        '--kafka-key'], help='Kafka API-Key (key)')
    context.argument('kafka_secret', options_list=[
        '--kafka-secret'], help='Kafka API-Key (secret)')
    context.argument('schema_registry', options_list=[
        '--schema-registry'], help='Schema registry url')
    context.argument('schema_key', options_list=[
        '--schema-key'], help='Schema registry API-Key (key)')
    context.argument('schema_secret', options_list=[
        '--schema-secret'], help='Schema registry API-Key (secret)')
    context.argument('connection_name', options_list=['--connection'],
                     help='Name of the connection', validator=validate_kafka_params)


def add_opt_out_argument(context):
    context.argument('opt_out_list', options_list=['--opt-out'],
                     default=None, nargs='+',
                     arg_type=get_enum_type(OPT_OUT_OPTION),
                     help='Whether to disable some configuration steps. '
                     'Use configinfo to disbale configuration information changes on source. '
                     'Use publicnetwork to disable public network access configuration.'
                     'Use auth to skip auth configuration such as enabling managed identity and granting RBAC roles.'
                     )


def load_arguments(self, _):  # pylint: disable=too-many-statements

    for source in SOURCE_RESOURCES_PARAMS:

        with self.argument_context('{} connection list'.format(source.value)) as c:
            add_source_resource_block(
                c, source, enable_id=False)

        with self.argument_context('{} connection show'.format(source.value)) as c:
            add_source_resource_block(c, source)
            add_connection_name_argument(c, source)

        with self.argument_context('{} connection delete'.format(source.value)) as c:
            add_connection_name_argument(c, source)
            add_source_resource_block(c, source)

        with self.argument_context('{} connection list-configuration'.format(source.value)) as c:
            add_connection_name_argument(c, source)
            add_source_resource_block(c, source)
        with self.argument_context('{} connection validate'.format(source.value)) as c:
            add_connection_name_argument(c, source)
            add_source_resource_block(c, source)

        with self.argument_context('{} connection list-support-types'.format(source.value)) as c:
            add_target_type_argument(c, source)

        with self.argument_context('{} connection wait'.format(source.value)) as c:
            add_connection_name_argument(c, source)
            add_source_resource_block(c, source)

        for target in TARGET_RESOURCES_PARAMS:
            with self.argument_context('{} connection create {}'.format(source.value, target.value)) as c:
                add_client_type_argument(c, source, target)
                add_connection_name_argument(c, source)
                add_source_resource_block(c, source, enable_id=False, target=target)
                add_target_resource_block(c, target)
                add_auth_block(c, source, target)
                add_new_addon_argument(c, source, target)
                add_configuration_store_argument(c)
                add_secret_store_argument(c, source)
                add_vnet_block(c, target)
                add_connection_string_argument(c, source, target)
                add_customized_keys_argument(c)
                add_opt_out_argument(c)
                add_connstr_props_argument(c)

            with self.argument_context('{} connection update {}'.format(source.value, target.value)) as c:
                add_client_type_argument(c, source, target)
                add_connection_name_argument(c, source)
                add_source_resource_block(c, source, enable_id=True, target=target)
                add_auth_block(c, source, target)
                add_configuration_store_argument(c)
                add_secret_store_argument(c, source)
                add_vnet_block(c, target)
                add_connection_string_argument(c, source, target)
                add_customized_keys_argument(c)
                add_opt_out_argument(c)
                add_connstr_props_argument(c)

        # special target resource: independent implementation
        target = RESOURCE.ConfluentKafka
        with self.argument_context('{} connection create {}'.format(source.value, target.value)) as c:
            add_client_type_argument(c, source, target)
            add_source_resource_block(c, source, enable_id=False)
            add_confluent_kafka_argument(c)
            add_configuration_store_argument(c)
            add_secret_store_argument(c, source)
            add_customized_keys_argument(c)
            add_opt_out_argument(c)
        with self.argument_context('{} connection update {}'.format(source.value, target.value)) as c:
            add_client_type_argument(c, source, target)
            add_source_resource_block(c, source, enable_id=False)
            add_confluent_kafka_argument(c)
            add_configuration_store_argument(c)
            add_secret_store_argument(c, source)
            add_customized_keys_argument(c)
            add_opt_out_argument(c)

    # local connection
    with self.argument_context('connection list') as c:
        add_local_connection_block(c, show_id=False)

    with self.argument_context('connection show') as c:
        add_local_connection_block(c)

    with self.argument_context('connection delete') as c:
        add_local_connection_block(c)

    with self.argument_context('connection generate-configuration') as c:
        add_local_connection_block(c)

    with self.argument_context('connection validate') as c:
        add_local_connection_block(c)

    with self.argument_context('connection list-support-types') as c:
        add_target_type_argument(c, source)

    with self.argument_context('connection wait') as c:
        add_local_connection_block(c)

    source = RESOURCE.Local
    for target in TARGET_RESOURCES_PARAMS:
        with self.argument_context('connection preview-configuration {}'.format(target.value)) as c:
            add_auth_block(c, source, target)
            add_client_type_argument(c, source, target)

        with self.argument_context('connection create {}'.format(target.value)) as c:
            add_client_type_argument(c, source, target)
            add_target_resource_block(c, target)
            add_auth_block(c, source, target)
            add_new_addon_argument(c, source, target)
            add_configuration_store_argument(c)
            add_secret_store_argument(c, source)
            add_vnet_block(c, target)
            add_local_connection_block(c)
            add_customized_keys_argument(c)
        with self.argument_context('connection update {}'.format(target.value)) as c:
            add_client_type_argument(c, source, target)
            add_auth_block(c, source, target)
            add_configuration_store_argument(c)
            add_secret_store_argument(c, source)
            add_vnet_block(c, target)
            add_local_connection_block(c)
            add_customized_keys_argument(c)

    # special target resource: independent implementation
    target = RESOURCE.ConfluentKafka
    with self.argument_context('connection create {}'.format(target.value)) as c:
        add_client_type_argument(c, source, target)
        add_confluent_kafka_argument(c)
        add_configuration_store_argument(c)
        add_secret_store_argument(c, source)
        add_local_connection_block(c, show_id=False)
        add_customized_keys_argument(c)
    with self.argument_context('connection update {}'.format(target.value)) as c:
        add_client_type_argument(c, source, target)
        add_confluent_kafka_argument(c)
        add_configuration_store_argument(c)
        add_secret_store_argument(c, source)
        add_local_connection_block(c, show_id=False)
        add_customized_keys_argument(c)
    with self.argument_context('connection preview-configuration {}'.format(target.value)) as c:
        add_auth_block(c, source, target)
        add_client_type_argument(c, source, target)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import json
from collections import OrderedDict
from urllib.parse import unquote

HEAD_PROPERTIES = {  # Convert response headers to properties.
    'Last-Modified': 'lastModified',
    'ocp-creation-time': 'creationTime',
    'ocp-batch-file-isdirectory': 'isDirectory',
    'ocp-batch-file-url': 'url',
    'ocp-batch-file-mode': 'fileMode',
    'Content-Length': 'contentLength',
    'Content-Type': 'contentType'
}


def _file_list_table_format(result):
    """Format file list as a table."""
    table_output = []
    for item in result:
        table_row = OrderedDict()
        table_row['Name'] = item.get('name', None)
        table_row['URL'] = item.get('url', None)
        table_row['Is Directory'] = str(item.get('isDirectory', None))
        table_row['Content Length'] = str(item.get('properties').get('contentLength', None)) \
            if item.get('properties', None) else ""
        table_row['Creation Time'] = item.get('properties').get('creationTime', None) \
            if item.get('properties', None) else ""
        table_output.append(table_row)
    return table_output


def _account_key_table_format(result):
    """Format account keys as a table."""
    table_output = []
    table_row = OrderedDict()
    table_row['Number'] = 'Primary'
    table_row['Key'] = result.get('primary', None)
    table_output.append(table_row)
    table_row = OrderedDict()
    table_row['Number'] = 'Secondary'
    table_row['Key'] = result.get('secondary', None)
    table_output.append(table_row)
    return table_output


def transform_response_headers(result):
    """Extract and format file property headers from ClientRawResponse object"""
    properties = {HEAD_PROPERTIES[k]: v for k, v in result.headers.items()
                  if k in HEAD_PROPERTIES}
    if properties.get('url'):
        properties['url'] = unquote(properties['url'])
    return properties


def task_file_list_table_format(result):
    """Format task file list as a table."""
    return _file_list_table_format(result)


def node_file_list_table_format(result):
    """Format node file list as a table."""
    return _file_list_table_format(result)


def application_list_table_format(result):
    """Format application list as a table."""
    table_output = []
    for item in result:
        table_row = OrderedDict()
        table_row['Id'] = item.get('id', None)
        table_row['Default Version'] = item.get('defaultVersion', None)
        table_row['Allow Updates'] = item.get('allowUpdates', None)
        table_row['Version Count'] = str(len(item['packages'])) if item.get('packages', None) else '0'
        table_output.append(table_row)
    return table_output


def application_summary_list_table_format(result):
    """Format application summary list as a table."""
    table_output = []
    for item in result:
        table_row = OrderedDict()
        table_row['Application Id'] = item.get('id', None)
        table_row['Display Name'] = item.get('displayName', None)
        table_row['Versions'] = json.dumps(item.get('versions', None))
        table_output.append(table_row)
    return table_output


def account_list_table_format(result):
    """Format account list as a table."""
    table_output = []
    for item in result:
        table_row = OrderedDict()
        table_row['Name'] = item.get('name', None)
        table_row['Location'] = item.get('location', None)
        table_row['Resource Group'] = item.get('resourceGroup', None)
        table_output.append(table_row)
    return table_output


def account_keys_list_table_format(result):
    """Format account keys list as a table."""
    return _account_key_table_format(result)


def account_keys_renew_table_format(result):
    """Format account keys renew as a table."""
    return _account_key_table_format(result)


def job_list_table_format(result):
    """Format job list as a table."""
    table_output = []
    for item in result:
        table_row = OrderedDict()
        table_row['Job Id'] = item.get('id', None)
        table_row['State'] = item.get('state', None)
        table_row['Previous State'] = item.get('previousState', None)
        table_row['Execution Pool'] = item.get('executionInfo').get('poolId', None) \
            if item.get('executionInfo', None) else ""
        table_output.append(table_row)
    return table_output


def job_prep_release_status_list_table_format(result):
    """Format job prep-release-status list as a table."""
    table_output = []
    for item in result:
        table_row = OrderedDict()
        table_row['Pool Id'] = item.get('poolId', None)
        table_row['Node Id'] = item.get('nodeId', None)
        table_row['Job Prep State'] = item.get('jobPreparationTaskExecutionInfo').get('state', None) \
            if item.get('jobPreparationTaskExecutionInfo', None) else ""
        table_row['Job Release State'] = item.get('jobReleaseTaskExecutionInfo').get('state', None) \
            if item.get('jobReleaseTaskExecutionInfo', None) else ""
        table_output.append(table_row)
    return table_output


def job_schedule_list_table_format(result):
    """Format job-schedule list as a table."""
    table_output = []
    for item in result:
        table_row = OrderedDict()
        table_row['Job Schedule Id'] = item.get('id', None)
        table_row['State'] = item.get('state', None)
        table_row['Previous State'] = item.get('previousState', None)
        table_output.append(table_row)
    return table_output


def node_list_table_format(result):
    """Format node list as a table."""
    table_output = []
    for item in result:
        table_row = OrderedDict()
        table_row['Node Id'] = item.get('id', None)
        table_row['State'] = item.get('state', None)
        table_row['VM Size'] = item.get('vmSize', None)
        table_row['IP Address'] = item.get('ipAddress', None)
        table_output.append(table_row)
    return table_output


def pool_node_agent_skus_list_table_format(result):
    """Format pool node-agent-skus list as a table."""
    table_output = []
    for item in result:
        table_row = OrderedDict()
        table_row['Agent Id'] = item.get('id', None)
        table_row['Publisher'] = item.get('publisher', None)
        table_row['Offer'] = item.get('offer', None)
        table_row['Sku'] = item.get('sku', None)
        table_output.append(table_row)
    return table_output


def pool_list_table_format(result):
    """Format pool list as a table."""
    table_output = []
    for item in result:
        table_row = OrderedDict()
        table_row['Pool Id'] = item.get('id', None)
        table_row['State'] = item.get('state', None)
        table_row['Allocation State'] = item.get('allocationState', None)
        table_row['VM Size'] = item.get('vmSize', None)
        table_row['Dedicated VM Count'] = item.get('currentDedicatedNodes', None)
        table_row['Low Priority VM Count'] = item.get('currentLowPriorityNodes', None)
        table_row['Type'] = 'IaaS' if item.get('virtualMachineConfiguration', None) else 'PaaS'
        table_output.append(table_row)
    return table_output


def pool_usage_metrics_list_table_format(result):
    """Format pool usage-metrics list as a table."""
    table_output = []
    for item in result:
        table_row = OrderedDict()
        table_row['Pool Id'] = item.get('poolId', None)
        table_row['Start Time'] = item.get('startTime') if item.get('startTime', None) else ""
        table_row['End Time'] = item.get('endTime') if item.get('endTime', None) else ""
        table_row['VM Size'] = item.get('vmSize', None)
        table_row['Total Core Hours'] = str(item.get('totalCoreHours', None))
        table_output.append(table_row)
    return table_output


def task_list_table_format(result):
    """Format task list as a table."""
    table_output = []
    for item in result:
        table_row = OrderedDict()
        table_row['Task Id'] = item.get('id', None)
        table_row['State'] = item.get('state', None)
        table_row['Exit Code'] = str(item.get('executionInfo').get('exitCode', None)) \
            if item.get('executionInfo', None) else ""
        table_row['Node Id'] = item.get('nodeInfo').get('nodeId', None) if item.get('nodeInfo', None) else ""
        table_row['Command Line'] = item.get('commandLine', None)
        table_output.append(table_row)
    return table_output


def task_create_table_format(result):
    """Format task create as a table."""
    table_output = []
    if not isinstance(result, list):
        table_row = OrderedDict()
        table_row['Task Id'] = result.get('id', None)
        table_row['Submission Status'] = "success"
        table_output.append(table_row)
    else:
        for item in result:
            table_row = OrderedDict()
            table_row['Task Id'] = item.get('taskId', None)
            table_row['Submission Status'] = item.get('status', None)
            table_row['Error'] = item.get('error').get('code', None) if item.get('error', None) else ""
            table_output.append(table_row)
    return table_output


def list_pool_node_counts_table_format(result):
    """Format account list pool node counts result as a table."""
    table_output = []
    for item in result:
        table_row = OrderedDict()
        table_row['Pool Id'] = item.get('poolId', None)
        table_row['Dedicated Starting'] = str(item.get('dedicated').get('starting', None)) \
            if item.get('dedicated', None) else ""
        table_row['Dedicated Idle'] = str(item.get('dedicated').get('idle', None)) \
            if item.get('dedicated', None) else ""
        table_row['Dedicated Running'] = str(item.get('dedicated').get('running', None)) \
            if item.get('dedicated', None) else ""
        table_row['Dedicated Total'] = str(item.get('dedicated').get('total', None)) \
            if item.get('dedicated', None) else ""
        table_row['LowPri Starting'] = str(item.get('lowPriority').get('starting', None)) \
            if item.get('lowPriority', None) else ""
        table_row['LowPri Idle'] = str(item.get('lowPriority').get('idle', None)) \
            if item.get('lowPriority', None) else ""
        table_row['LowPri Running'] = str(item.get('lowPriority').get('running', None)) \
            if item.get('lowPriority', None) else ""
        table_row['LowPri Total'] = str(item.get('lowPriority').get('total', None)) \
            if item.get('lowPriority', None) else ""
        table_output.append(table_row)
    return table_output


def list_supported_images_table_format(result):
    """Format account list node agent skus result as a table."""
    table_output = []
    for item in result:
        table_row = OrderedDict()
        table_row['OS Type'] = item.get('osType', None)
        table_row['Node Agent Sku'] = item.get('nodeAgentSkuId', None)
        table_row['Publisher'] = item.get('imageReference').get('publisher', None) \
            if item.get('imageReference', None) else ""
        table_row['Offer'] = item.get('imageReference').get('offer', None) \
            if item.get('imageReference', None) else ""
        table_row['Sku'] = item.get('imageReference').get('sku', None) \
            if item.get('imageReference', None) else ""
        table_row['Version'] = item.get('imageReference').get('version', None) \
            if item.get('imageReference', None) else ""
        table_row['VerificationType'] = item.get('verificationType', None)
        table_output.append(table_row)
    return table_output



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import collections
import random
from base64 import b64decode
import textwrap

from azure.core.exceptions import HttpResponseError, \
    ResourceNotFoundError as CoreResourceNotFoundError
from azure.mgmt.core.tools import resource_id, parse_resource_id
import azure.mgmt.redhatopenshift.models as openshiftcluster

from azure.cli.core.commands.client_factory import get_mgmt_service_client
from azure.cli.core.commands.client_factory import get_subscription_id
from azure.cli.core.profiles import ResourceType
from azure.cli.core.util import sdk_no_wait
from azure.cli.core.azclierror import FileOperationError, ResourceNotFoundError, \
    UnauthorizedError, ValidationError
from azure.cli.command_modules.aro._aad import AADManager
from azure.cli.command_modules.aro._rbac import assign_role_to_resource, \
    has_role_assignment_on_resource
from azure.cli.command_modules.aro._rbac import ROLE_NETWORK_CONTRIBUTOR, ROLE_READER
from azure.cli.command_modules.aro._validators import validate_subnets
from azure.cli.command_modules.aro._dynamic_validators import validate_cluster_create
from azure.cli.command_modules.aro.aaz.latest.network.vnet.subnet import Show as subnet_show
from azure.cli.command_modules.role import GraphError

from knack.log import get_logger

from msrest.exceptions import HttpOperationError

from tabulate import tabulate

logger = get_logger(__name__)

FP_CLIENT_ID = 'f1dd0a37-89c6-4e07-bcd1-ffd3d43d8875'


def aro_create(cmd,  # pylint: disable=too-many-locals
               client,
               resource_group_name,
               resource_name,
               master_subnet,
               worker_subnet,
               vnet=None,  # pylint: disable=unused-argument
               vnet_resource_group_name=None,  # pylint: disable=unused-argument
               enable_preconfigured_nsg=False,
               location=None,
               pull_secret=None,
               domain=None,
               cluster_resource_group=None,
               fips_validated_modules=False,
               client_id=None,
               client_secret=None,
               pod_cidr='10.128.0.0/14',
               service_cidr='172.30.0.0/16',
               outbound_type='Loadbalancer',
               disk_encryption_set=None,
               master_encryption_at_host=False,
               master_vm_size='Standard_D8s_v5',
               worker_encryption_at_host=False,
               worker_vm_size='Standard_D4s_v5',
               worker_vm_disk_size_gb='128',
               worker_count='3',
               apiserver_visibility='Public',
               ingress_visibility='Public',
               load_balancer_managed_outbound_ip_count=None,
               tags=None,
               version=None,
               no_wait=False):

    resource_client = get_mgmt_service_client(
        cmd.cli_ctx, ResourceType.MGMT_RESOURCE_RESOURCES)
    provider = resource_client.providers.get('Microsoft.RedHatOpenShift')
    if provider.registration_state != 'Registered':
        raise UnauthorizedError('Microsoft.RedHatOpenShift provider is not registered.',
                                'Run `az provider register -n Microsoft.RedHatOpenShift --wait`.')

    validate_subnets(master_subnet, worker_subnet)

    validate(cmd,
             client,
             resource_group_name,
             resource_name,
             master_subnet,
             worker_subnet,
             vnet=vnet,
             enable_preconfigured_nsg=enable_preconfigured_nsg,
             cluster_resource_group=cluster_resource_group,
             client_id=client_id,
             client_secret=client_secret,
             vnet_resource_group_name=vnet_resource_group_name,
             disk_encryption_set=disk_encryption_set,
             location=location,
             version=version,
             pod_cidr=pod_cidr,
             service_cidr=service_cidr,
             warnings_as_text=True)

    subscription_id = get_subscription_id(cmd.cli_ctx)

    random_id = generate_random_id()

    aad = AADManager(cmd.cli_ctx)
    if client_id is None:
        client_id, client_secret = aad.create_application(cluster_resource_group or 'aro-' + random_id)

    client_sp_id = aad.get_service_principal_id(client_id)
    if not client_sp_id:
        client_sp_id = aad.create_service_principal(client_id)

    rp_client_sp_id = aad.get_service_principal_id(resolve_rp_client_id())
    if not rp_client_sp_id:
        raise ResourceNotFoundError("RP service principal not found.")

    worker_vm_size = worker_vm_size or 'Standard_D4s_v5'

    if apiserver_visibility is not None:
        apiserver_visibility = apiserver_visibility.capitalize()

    if ingress_visibility is not None:
        ingress_visibility = ingress_visibility.capitalize()

    load_balancer_profile = None
    if load_balancer_managed_outbound_ip_count is not None:
        load_balancer_profile = openshiftcluster.LoadBalancerProfile()
        load_balancer_profile.managed_outbound_ips = openshiftcluster.ManagedOutboundIPs()
        load_balancer_profile.managed_outbound_ips.count = load_balancer_managed_outbound_ip_count  # pylint: disable=line-too-long

    oc = openshiftcluster.OpenShiftCluster(
        location=location,
        tags=tags,
        cluster_profile=openshiftcluster.ClusterProfile(
            pull_secret=pull_secret or "",
            domain=domain or random_id,
            resource_group_id=(f"/subscriptions/{subscription_id}"
                               f"/resourceGroups/{cluster_resource_group or 'aro-' + random_id}"),
            fips_validated_modules='Enabled' if fips_validated_modules else 'Disabled',
            version=version or '',
        ),
        service_principal_profile=openshiftcluster.ServicePrincipalProfile(
            client_id=client_id,
            client_secret=client_secret,
        ),
        network_profile=openshiftcluster.NetworkProfile(
            pod_cidr=pod_cidr or '10.128.0.0/14',
            service_cidr=service_cidr or '172.30.0.0/16',
            outbound_type=outbound_type or '',
            load_balancer_profile=load_balancer_profile,
            preconfigured_nsg='Enabled' if enable_preconfigured_nsg else 'Disabled',
        ),
        master_profile=openshiftcluster.MasterProfile(
            vm_size=master_vm_size or 'Standard_D8s_v5',
            subnet_id=master_subnet,
            encryption_at_host='Enabled' if master_encryption_at_host else 'Disabled',
            disk_encryption_set_id=disk_encryption_set,
        ),
        worker_profiles=[
            openshiftcluster.WorkerProfile(
                name='worker',  # TODO: 'worker' should not be hard-coded
                vm_size=worker_vm_size,
                disk_size_gb=worker_vm_disk_size_gb or 128,
                subnet_id=worker_subnet,
                count=worker_count or 3,
                encryption_at_host='Enabled' if worker_encryption_at_host else 'Disabled',
                disk_encryption_set_id=disk_encryption_set,
            )
        ],
        apiserver_profile=openshiftcluster.APIServerProfile(
            visibility=apiserver_visibility or 'Public',
        ),
        ingress_profiles=[
            openshiftcluster.IngressProfile(
                name='default',  # TODO: 'default' should not be hard-coded
                visibility=ingress_visibility or 'Public',
            )
        ],
    )

    sp_obj_ids = [client_sp_id, rp_client_sp_id]
    ensure_resource_permissions(cmd.cli_ctx, oc, True, sp_obj_ids)

    return sdk_no_wait(no_wait, client.open_shift_clusters.begin_create_or_update,
                       resource_group_name=resource_group_name,
                       resource_name=resource_name,
                       parameters=oc)


def validate(cmd,  # pylint: disable=too-many-locals,too-many-statements
             client,  # pylint: disable=unused-argument
             resource_group_name,  # pylint: disable=unused-argument
             resource_name,  # pylint: disable=unused-argument
             master_subnet,
             worker_subnet,
             vnet=None,
             enable_preconfigured_nsg=None,
             cluster_resource_group=None,  # pylint: disable=unused-argument
             client_id=None,
             client_secret=None,  # pylint: disable=unused-argument
             vnet_resource_group_name=None,  # pylint: disable=unused-argument
             disk_encryption_set=None,
             location=None,  # pylint: disable=unused-argument
             version=None,
             pod_cidr=None,  # pylint: disable=unused-argument
             service_cidr=None,  # pylint: disable=unused-argument
             warnings_as_text=False):

    class mockoc:  # pylint: disable=too-few-public-methods
        def __init__(self, disk_encryption_id, master_subnet_id, worker_subnet_id, preconfigured_nsg):
            self.network_profile = openshiftcluster.NetworkProfile(
                preconfigured_nsg='Enabled' if preconfigured_nsg else 'Disabled'
            )
            self.master_profile = openshiftcluster.MasterProfile(
                subnet_id=master_subnet_id,
                disk_encryption_set_id=disk_encryption_id
            )
            self.worker_profiles = [openshiftcluster.WorkerProfile(
                subnet_id=worker_subnet_id
            )]
            self.worker_profiles_status = None

    aad = AADManager(cmd.cli_ctx)

    rp_client_sp_id = aad.get_service_principal_id(resolve_rp_client_id())
    if not rp_client_sp_id:
        raise ResourceNotFoundError("RP service principal not found.")

    sp_obj_ids = [rp_client_sp_id]

    if client_id is not None:
        sp_obj_ids.append(aad.get_service_principal_id(client_id))

    cluster = mockoc(disk_encryption_set, master_subnet, worker_subnet, enable_preconfigured_nsg)
    try:
        # Get cluster resources we need to assign permissions on, sort to ensure the same order of operations
        resources = {ROLE_NETWORK_CONTRIBUTOR: sorted(get_cluster_network_resources(cmd.cli_ctx, cluster, True)),
                     ROLE_READER: sorted(get_disk_encryption_resources(cluster))}
    except (HttpResponseError, HttpOperationError) as e:
        logger.error(e.message)
        raise

    if vnet is None:
        master_parts = parse_resource_id(master_subnet)
        vnet = resource_id(
            subscription=master_parts['subscription'],
            resource_group=master_parts['resource_group'],
            namespace='Microsoft.Network',
            type='virtualNetworks',
            name=master_parts['name'],
        )

    error_objects = validate_cluster_create(version,
                                            resources,
                                            sp_obj_ids)
    errors_and_warnings = []
    for error_func in error_objects:
        namespace = collections.namedtuple("Namespace", locals().keys())(*locals().values())
        error_obj = error_func(cmd, namespace)
        if error_obj != []:
            for err in error_obj:
                # Wrap text so tabulate returns a pretty table
                new_err = []
                for txt in err:
                    new_err.append(textwrap.fill(txt, width=160))
                errors_and_warnings.append(new_err)

    warnings = []
    errors = []
    if len(errors_and_warnings) > 0:
        # Separate errors and warnings into separate arrays
        for issue in errors_and_warnings:
            if issue[2] == "Warning":
                warnings.append(issue)
            else:
                errors.append(issue)
    else:
        logger.info("No validation errors or warnings")

    if len(warnings) > 0:
        if len(errors) == 0 and warnings_as_text:
            full_msg = ""
            for warning in warnings:
                full_msg = full_msg + f"{warning[3]}\n"
        else:
            headers = ["Type", "Name", "Severity", "Description"]
            table = tabulate(warnings, headers=headers, tablefmt="grid")
            full_msg = f"The following issues will have a minor impact on cluster creation:\n{table}"
        logger.warning(full_msg)

    if len(errors) > 0:
        if len(warnings) > 0:
            full_msg = "\n"
        else:
            full_msg = ""
        headers = ["Type", "Name", "Severity", "Description"]
        table = tabulate(errors, headers=headers, tablefmt="grid")
        full_msg = full_msg + f"The following errors are fatal and will block cluster creation:\n{table}"
        raise ValidationError(full_msg)


def aro_validate(cmd,  # pylint: disable=too-many-locals,too-many-statements
                 client,  # pylint: disable=unused-argument
                 resource_group_name,  # pylint: disable=unused-argument
                 resource_name,  # pylint: disable=unused-argument
                 master_subnet,
                 worker_subnet,
                 vnet=None,
                 cluster_resource_group=None,  # pylint: disable=unused-argument
                 client_id=None,
                 client_secret=None,  # pylint: disable=unused-argument
                 vnet_resource_group_name=None,  # pylint: disable=unused-argument
                 disk_encryption_set=None,
                 location=None,  # pylint: disable=unused-argument
                 version=None,
                 pod_cidr=None,  # pylint: disable=unused-argument
                 service_cidr=None,  # pylint: disable=unused-argument
                 ):

    validate(cmd,
             client,
             resource_group_name,
             resource_name,
             master_subnet,
             worker_subnet,
             vnet=vnet,
             cluster_resource_group=cluster_resource_group,
             client_id=client_id,
             client_secret=client_secret,
             vnet_resource_group_name=vnet_resource_group_name,
             disk_encryption_set=disk_encryption_set,
             location=location,
             version=version,
             pod_cidr=pod_cidr,
             service_cidr=service_cidr,
             warnings_as_text=False)


def aro_delete(cmd, client, resource_group_name, resource_name, no_wait=False):
    # TODO: clean up rbac
    rp_client_sp_id = None

    try:
        oc = client.open_shift_clusters.get(resource_group_name, resource_name)
    except HttpResponseError as e:
        if e.status_code == 404:
            raise ResourceNotFoundError(e.message) from e
        logger.info(e.message)
    except HttpOperationError as e:
        logger.info(e.message)

    aad = AADManager(cmd.cli_ctx)

    # Best effort - assume the role assignments on the SP exist if exception raised
    try:
        rp_client_sp_id = aad.get_service_principal_id(resolve_rp_client_id())
        if not rp_client_sp_id:
            raise ResourceNotFoundError("RP service principal not found.")
    except GraphError as e:
        logger.info(e)

    # Customers frequently remove the Cluster or RP's service principal permissions.
    # Attempt to fix this before performing any action against the cluster
    if rp_client_sp_id:
        ensure_resource_permissions(cmd.cli_ctx, oc, False, [rp_client_sp_id])

    return sdk_no_wait(no_wait, client.open_shift_clusters.begin_delete,
                       resource_group_name=resource_group_name,
                       resource_name=resource_name)


def aro_list(client, resource_group_name=None):
    if resource_group_name:
        return client.open_shift_clusters.list_by_resource_group(resource_group_name)
    return client.open_shift_clusters.list()


def aro_show(client, resource_group_name, resource_name):
    return client.open_shift_clusters.get(resource_group_name, resource_name)


def aro_list_credentials(client, resource_group_name, resource_name):
    return client.open_shift_clusters.list_credentials(resource_group_name, resource_name)


def aro_get_admin_kubeconfig(client, resource_group_name, resource_name, file="kubeconfig"):
    query_result = client.open_shift_clusters.list_admin_credentials(resource_group_name, resource_name)
    file_mode = "x"
    yaml_data = b64decode(query_result.kubeconfig).decode('UTF-8')
    try:
        with open(file, file_mode, encoding="utf-8") as f:
            f.write(yaml_data)
    except FileExistsError as e:
        raise FileOperationError(f"File {file} already exists.") from e
    logger.info("Kubeconfig written to file: %s", file)


def aro_get_versions(client, location):
    items = client.open_shift_versions.list(location)
    versions = []
    for item in items:
        versions.append(item.version)
    return sorted(versions)


def aro_update(cmd,
               client,
               resource_group_name,
               resource_name,
               refresh_cluster_credentials=False,
               client_id=None,
               client_secret=None,
               load_balancer_managed_outbound_ip_count=None,
               no_wait=False):
    # if we can't read cluster spec, we will not be able to do much. Fail.
    oc = client.open_shift_clusters.get(resource_group_name, resource_name)

    ocUpdate = openshiftcluster.OpenShiftClusterUpdate()

    client_id, client_secret = cluster_application_update(cmd.cli_ctx, oc, client_id, client_secret, refresh_cluster_credentials)  # pylint: disable=line-too-long

    if client_id is not None or client_secret is not None:
        # construct update payload
        ocUpdate.service_principal_profile = openshiftcluster.ServicePrincipalProfile()

        if client_secret is not None:
            ocUpdate.service_principal_profile.client_secret = client_secret

        if client_id is not None:
            ocUpdate.service_principal_profile.client_id = client_id

    if load_balancer_managed_outbound_ip_count is not None:
        ocUpdate.network_profile = openshiftcluster.NetworkProfile()
        ocUpdate.network_profile.load_balancer_profile = openshiftcluster.LoadBalancerProfile()
        ocUpdate.network_profile.load_balancer_profile.managed_outbound_ips = openshiftcluster.ManagedOutboundIPs()
        ocUpdate.network_profile.load_balancer_profile.managed_outbound_ips.count = load_balancer_managed_outbound_ip_count  # pylint: disable=line-too-long

    return sdk_no_wait(no_wait, client.open_shift_clusters.begin_update,
                       resource_group_name=resource_group_name,
                       resource_name=resource_name,
                       parameters=ocUpdate)


def generate_random_id():
    random_id = (random.choice('abcdefghijklmnopqrstuvwxyz') +
                 ''.join(random.choice('abcdefghijklmnopqrstuvwxyz1234567890')
                         for _ in range(7)))
    return random_id


def get_network_resources_from_subnets(cli_ctx, subnets, fail, oc):

    subnet_resources = set()
    subnets_with_no_nsg_attached = set()
    for sn in subnets:
        sid = parse_resource_id(sn)

        if 'resource_group' not in sid or 'name' not in sid or 'resource_name' not in sid:
            if fail:
                raise ValidationError(f"""(ValidationError) Failed to validate subnet '{sn}'.
                    Please retry, if issue persists: raise azure support ticket""")
            logger.info("Failed to validate subnet '%s'", sn)

        try:
            subnet = subnet_show(cli_ctx=cli_ctx)(command_args={
                "name": sid['resource_name'],
                "vnet_name": sid['name'],
                "resource_group": sid['resource_group']}
            )
        except CoreResourceNotFoundError:
            continue

        if subnet.get("routeTable", None):
            subnet_resources.add(subnet["routeTable"]["id"])

        if subnet.get("natGateway", None):
            subnet_resources.add(subnet["natGateway"]["id"])

        if oc.network_profile.preconfigured_nsg == 'Enabled':
            if subnet.get("networkSecurityGroup", None):
                subnet_resources.add(subnet['networkSecurityGroup']['id'])
            else:
                subnets_with_no_nsg_attached.add(sn)

    # when preconfiguredNSG feature is Enabled we either have all subnets NSG attached or none.
    if oc.network_profile.preconfigured_nsg == 'Enabled' and \
        len(subnets_with_no_nsg_attached) != 0 and \
            len(subnets_with_no_nsg_attached) != len(subnets):
        raise ValidationError(f"(ValidationError) preconfiguredNSG feature is enabled but an NSG is\
                               not attached for all required subnets. Please make sure all the following\
                               subnets have a network security groups attached and retry.\
                              {subnets_with_no_nsg_attached}")

    return subnet_resources


def get_cluster_network_resources(cli_ctx, oc, fail):
    master_subnet = oc.master_profile.subnet_id
    worker_subnets = set()

    # Ensure that worker_profiles exists
    # it will not be returned if the cluster resources do not exist
    if oc.worker_profiles is not None:
        worker_subnets = {w.subnet_id for w in oc.worker_profiles}

    # Ensure that worker_profiles_status exists
    # it will not be returned if the cluster resources do not exist

    # We filter nonexistent subnets here as we only propagate subnet values for
    # worker profiles/machinesets considered valid.
    if oc.worker_profiles_status is not None:
        worker_subnets |= {w.subnet_id for w in oc.worker_profiles_status if w.subnet_id is not None}

    master_parts = parse_resource_id(master_subnet)
    vnet = resource_id(
        subscription=master_parts['subscription'],
        resource_group=master_parts['resource_group'],
        namespace='Microsoft.Network',
        type='virtualNetworks',
        name=master_parts['name'],
    )

    return get_network_resources(cli_ctx, worker_subnets | {master_subnet}, vnet, fail, oc)


def get_network_resources(cli_ctx, subnets, vnet, fail, oc):
    subnet_resources = get_network_resources_from_subnets(cli_ctx, subnets, fail, oc)

    resources = set()
    resources.add(vnet)
    resources.update(subnet_resources)

    return resources


def get_disk_encryption_resources(oc):
    disk_encryption_set = oc.master_profile.disk_encryption_set_id
    resources = set()
    if disk_encryption_set:
        resources.add(disk_encryption_set)
    return resources


# cluster_application_update manages cluster application & service principal update
# If called without parameters it should be best-effort
# If called with parameters it fails if something is not possible
# Flow:
# 1. Set fail - if we are in fail mode or best effort.
# 2. Sort out client_id, rp_client_sp, resources we care for RBAC.
# 3. If we are in refresh_cluster_credentials mode - attempt to reuse/recreate
# cluster service principal application and acquire client_id, client_secret
# 4. Reuse/Recreate service principal.
# 5. Sort out required rbac
def cluster_application_update(cli_ctx,
                               oc,
                               client_id,
                               client_secret,
                               refresh_cluster_credentials):
    # QUESTION: is there possible unification with the create path?

    rp_client_sp_id = None
    client_sp_id = None
    random_id = generate_random_id()

    # if any of these are set - we expect users to have access to fix rbac so we fail
    # common for 1 and 2 flows
    fail = client_id or client_secret or refresh_cluster_credentials

    aad = AADManager(cli_ctx)

    # check if we can see if RP service principal exists
    try:
        rp_client_sp_id = aad.get_service_principal_id(resolve_rp_client_id())
        if not rp_client_sp_id:
            raise ResourceNotFoundError("RP service principal not found.")
    except GraphError as e:
        if fail:
            logger.error(e)
            raise
        logger.info(e)

    # refresh_cluster_credentials refreshes cluster SP application.
    # At firsts it tries to re-use existing application and generate new password.
    # If application does not exist - creates new one
    if refresh_cluster_credentials:
        try:
            app = aad.get_application_object_id_by_client_id(client_id or oc.service_principal_profile.client_id)
            if not app:
                # we were not able to find and applications, create new one
                parts = parse_resource_id(oc.cluster_profile.resource_group_id)
                cluster_resource_group = parts['resource_group']

                client_id, client_secret = aad.create_application(cluster_resource_group or 'aro-' + random_id)
            else:
                client_secret = aad.add_password(app)
        except GraphError as e:
            logger.error(e)
            raise

    # attempt to get/create SP if one was not found.
    try:
        client_sp_id = aad.get_service_principal_id(client_id or oc.service_principal_profile.client_id)
    except GraphError as e:
        if fail:
            logger.error(e)
            raise
        logger.info(e)

    if fail and not client_sp_id:
        client_sp_id = aad.create_service_principal(client_id or oc.service_principal_profile.client_id)

    sp_obj_ids = [sp for sp in [rp_client_sp_id, client_sp_id] if sp]
    ensure_resource_permissions(cli_ctx, oc, fail, sp_obj_ids)

    return client_id, client_secret


def resolve_rp_client_id():
    return FP_CLIENT_ID


def ensure_resource_permissions(cli_ctx, oc, fail, sp_obj_ids):
    try:
        # Get cluster resources we need to assign permissions on, sort to ensure the same order of operations
        resources = {ROLE_NETWORK_CONTRIBUTOR: sorted(get_cluster_network_resources(cli_ctx, oc, fail)),
                     ROLE_READER: sorted(get_disk_encryption_resources(oc))}
    except (HttpResponseError, HttpOperationError) as e:
        if fail:
            logger.error(e.message)
            raise
        logger.info(e.message)
        return

    for sp_id in sp_obj_ids:
        for role in sorted(resources):
            for resource in resources[role]:
                # Create the role assignment if it doesn't exist
                # Assume that the role assignment exists if we fail to look it up
                resource_contributor_exists = True
                try:
                    resource_contributor_exists = has_role_assignment_on_resource(cli_ctx, resource, sp_id, role)
                except HttpResponseError as e:
                    if fail:
                        logger.error(e.message)
                        raise
                    logger.info(e.message)

                if not resource_contributor_exists:
                    assign_role_to_resource(cli_ctx, resource, sp_id, role)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import uuid

from azure.cli.core.commands.client_factory import get_mgmt_service_client
from azure.cli.core.commands.client_factory import get_subscription_id
from azure.cli.core.profiles import get_sdk
from azure.cli.core.profiles import ResourceType
from azure.mgmt.core.tools import resource_id
from knack.log import get_logger
from msrest.exceptions import ValidationError

ROLE_NETWORK_CONTRIBUTOR = '4d97b98b-1d4f-4787-a291-c67834d212e7'
ROLE_READER = 'acdd72a7-3385-48ef-bd42-f606fba81ae7'

logger = get_logger(__name__)


def _gen_uuid():
    return uuid.uuid4()


def _create_role_assignment(auth_client, resource, params):
    # retry "ValidationError: A hash conflict was encountered for the role Assignment ID. Please use a new Guid."
    max_retries = 3
    retries = 0
    while True:
        try:
            return auth_client.role_assignments.create(resource, _gen_uuid(), params)
        except ValidationError as ex:
            if retries >= max_retries:
                raise
            retries += 1
            logger.warning("%s; retry %d of %d", ex, retries, max_retries)


def assign_role_to_resource(cli_ctx, resource, object_id, role_name):
    auth_client = get_mgmt_service_client(cli_ctx, ResourceType.MGMT_AUTHORIZATION)

    RoleAssignmentCreateParameters = get_sdk(cli_ctx, ResourceType.MGMT_AUTHORIZATION,
                                             'RoleAssignmentCreateParameters', mod='models',
                                             operation_group='role_assignments')

    role_definition_id = resource_id(
        subscription=get_subscription_id(cli_ctx),
        namespace='Microsoft.Authorization',
        type='roleDefinitions',
        name=role_name,
    )

    _create_role_assignment(auth_client, resource, RoleAssignmentCreateParameters(
        role_definition_id=role_definition_id,
        principal_id=object_id,
        principal_type='ServicePrincipal',
    ))


def has_role_assignment_on_resource(cli_ctx, resource, object_id, role_name):
    auth_client = get_mgmt_service_client(cli_ctx, ResourceType.MGMT_AUTHORIZATION)

    role_definition_id = resource_id(
        subscription=get_subscription_id(cli_ctx),
        namespace='Microsoft.Authorization',
        type='roleDefinitions',
        name=role_name,
    )

    for assignment in auth_client.role_assignments.list_for_scope(resource):
        if assignment.role_definition_id.lower() == role_definition_id.lower() and \
                assignment.principal_id.lower() == object_id.lower():
            return True

    return False



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
# pylint: disable=line-too-long
import uuid
import os

from knack.log import get_logger

from azure.cli.core.commands import LongRunningOperation

from azure.cli.command_modules.vm.custom import _is_linux_os_aaz
from azure.cli.command_modules.vm._vm_utils import get_key_vault_base_url, create_data_plane_keyvault_key_client

_DATA_VOLUME_TYPE = 'DATA'
_ALL_VOLUME_TYPE = 'ALL'
_STATUS_ENCRYPTED = 'Encrypted'

logger = get_logger(__name__)

vm_extension_info = {
    'Linux': {
        'publisher': os.environ.get('ADE_TEST_EXTENSION_PUBLISHER') or 'Microsoft.Azure.Security',
        'name': os.environ.get('ADE_TEST_EXTENSION_NAME') or 'AzureDiskEncryptionForLinux',
        'version': '1.1',
        'legacy_version': '0.1'
    },
    'Windows': {
        'publisher': os.environ.get('ADE_TEST_EXTENSION_PUBLISHER') or 'Microsoft.Azure.Security',
        'name': os.environ.get('ADE_TEST_EXTENSION_NAME') or 'AzureDiskEncryption',
        'version': '2.2',
        'legacy_version': '1.1'
    }
}


def _find_existing_ade(vm, use_instance_view=False, ade_ext_info=None):
    if not ade_ext_info:
        ade_ext_info = vm_extension_info['Linux'] if _is_linux_os_aaz(vm) else vm_extension_info['Windows']
    if use_instance_view:
        exts = vm['instanceView'].get('extensions', [])
        r = next((e for e in exts if e['type'].lower().startswith(ade_ext_info['publisher'].lower()) and
                  e['name'].lower() == ade_ext_info['name'].lower()), None)
    else:
        exts = vm.get('resources', [])
        r = next((e for e in exts if (e['publisher'].lower() == ade_ext_info['publisher'].lower() and
                                      e['typePropertiesType'].lower() == ade_ext_info['name'].lower())), None)
    return r


def _detect_ade_status(vm):
    if vm.get('storageProfile', {}).get('osDisk', {}).get('encryptionSettings', []):
        return False, True
    ade_ext_info = vm_extension_info['Linux'] if _is_linux_os_aaz(vm) else vm_extension_info['Windows']
    ade = _find_existing_ade(vm, ade_ext_info=ade_ext_info)
    if ade is None:
        return False, False
    if ade['typeHandlerVersion'].split('.')[0] == ade_ext_info['legacy_version'].split('.', maxsplit=1)[0]:
        return False, True

    return True, False   # we believe impossible to have both old & new ADE


def updateVmEncryptionSetting(cmd, vm, resource_group_name, vm_name, encryption_identity):
    from azure.cli.core.azclierror import ArgumentUsageError
    if encryption_identity.lower() not in (k.lower() for k in vm.get('identity', {}).get('userAssignedIdentities', {}).keys()):
        raise ArgumentUsageError("Encryption Identity should be an ARM Resource ID of one of the "
                                 "user assigned identities associated to the resource")

    updateVm = False

    if not (_encrypt_userid := vm.get('securityProfile', {}).get('encryptionIdentity', {}).get('userAssignedIdentityResourceId', None)) \
       or _encrypt_userid.lower() != encryption_identity.lower():
        updateVm = True

    if updateVm:
        from .aaz.latest.vm import Patch as VMPatchUpdate
        security_profile = {
            'encryption_identity': {
                'user_assigned_identity_resource_id': encryption_identity
            }
        }
        updateEncryptionIdentity = VMPatchUpdate(cli_ctx=cmd.cli_ctx)(command_args={
            'location': vm['location'],
            'vm_name': vm_name,
            'resource_group': resource_group_name,
            'security_profile': security_profile
        })
        result = LongRunningOperation(cmd.cli_ctx)(updateEncryptionIdentity)
        return result is not None and result['provisioningState'] == 'Succeeded'
    logger.info("No changes in identity")
    return True


def updateVmssEncryptionSetting(cmd, vmss, resource_group_name, vmss_name, encryption_identity):
    from azure.cli.core.azclierror import ArgumentUsageError
    if encryption_identity.lower() not in (k.lower() for k in vmss.get('identity', {}).get('userAssignedIdentities', {}).keys()):
        raise ArgumentUsageError("Encryption Identity should be an ARM Resource ID of one of the "
                                 "user assigned identities associated to the resource")

    updateVmss = False
    if vmss['properties']['virtualMachineProfile'].get('securityProfile', {}).get('encryptionIdentity', {}).\
            get('userAssignedIdentityResourceId', '').lower() != encryption_identity.lower():
        updateVmss = True

    if updateVmss:
        from .aaz.latest.vmss import Patch
        virtual_machine_profile = {
            'securityProfile': {
                'encryptionIdentity': {
                    'userAssignedIdentityResourceId': encryption_identity
                }
            }
        }
        updateEncryptionIdentity = Patch(cli_ctx=cmd.cli_ctx)(command_args={
            'vm_scale_set_name': vmss_name,
            'resource_group': resource_group_name,
            'virtual_machine_profile': virtual_machine_profile
        })
        LongRunningOperation(cmd.cli_ctx)(updateEncryptionIdentity)
        result = updateEncryptionIdentity.result()
        return result is not None and result.provisioning_state == 'Succeeded'
    logger.info("No changes in identity")
    return True


def isVersionSuppprtedForEncryptionIdentity(cmd):
    from azure.cli.core.profiles import ResourceType
    from knack.util import CLIError
    if not cmd.supported_api_version(min_api='2023-09-01', resource_type=ResourceType.MGMT_COMPUTE):
        raise CLIError("Usage error: Encryption Identity required API version 2023-09-01 or higher."
                       "You can set the cloud's profile to use the required API Version with:"
                       "az cloud set --profile latest --name <cloud name>")
    return True


def encrypt_vm(cmd, resource_group_name, vm_name,  # pylint: disable=too-many-locals, too-many-statements
               disk_encryption_keyvault,
               aad_client_id=None,
               aad_client_secret=None, aad_client_cert_thumbprint=None,
               key_encryption_keyvault=None,
               key_encryption_key=None,
               key_encryption_algorithm='RSA-OAEP',
               volume_type=None,
               encrypt_format_all=False,
               force=False, encryption_identity=None):
    from azure.mgmt.core.tools import parse_resource_id
    from knack.util import CLIError

    from .operations.vm import VMShow
    vm = VMShow(cli_ctx=cmd.cli_ctx)(command_args={
        'vm_name': vm_name,
        'resource_group': resource_group_name
    })
    is_linux = _is_linux_os_aaz(vm)
    vm_encrypted = bool(vm['storageProfile']['osDisk'].get('encryptionSettings', {}).get('enabled', False))
    _, has_old_ade = _detect_ade_status(vm)

    use_new_ade = not aad_client_id and not has_old_ade
    extension = vm_extension_info['Linux' if is_linux else 'Windows']

    if not use_new_ade and not aad_client_id:
        raise CLIError('Please provide --aad-client-id')

    # 1. First validate arguments
    if not use_new_ade and not aad_client_cert_thumbprint and not aad_client_secret:
        raise CLIError('Please provide either --aad-client-cert-thumbprint or --aad-client-secret')

    if volume_type is None:
        if not is_linux:
            volume_type = _ALL_VOLUME_TYPE
        elif vm['storageProfile'].get('dataDisks', []):
            raise CLIError('VM has data disks, please supply --volume-type')
        else:
            volume_type = 'OS'

    # sequence_version should be unique
    sequence_version = uuid.uuid4()

    # retrieve keyvault details
    disk_encryption_keyvault_url = get_key_vault_base_url(
        cmd.cli_ctx, (parse_resource_id(disk_encryption_keyvault))['name'])

    # disk encryption key itself can be further protected, so let us verify
    if key_encryption_key:
        key_encryption_keyvault = key_encryption_keyvault or disk_encryption_keyvault
    if encryption_identity and isVersionSuppprtedForEncryptionIdentity(cmd):
        result = updateVmEncryptionSetting(cmd, vm, resource_group_name, vm_name, encryption_identity)
        if result:
            logger.info("Encryption Identity successfully set in virtual machine")
        else:
            raise CLIError("Failed to update encryption Identity to the VM")

    #  to avoid bad server errors, ensure the vault has the right configurations
    _verify_keyvault_good_for_encryption(cmd.cli_ctx, disk_encryption_keyvault, key_encryption_keyvault, vm, force)

    # if key name and not key url, get url.
    if key_encryption_key and '://' not in key_encryption_key:  # if key name and not key url
        key_encryption_key = _get_keyvault_key_url(
            cmd.cli_ctx, (parse_resource_id(key_encryption_keyvault))['name'], key_encryption_key)

    # 2. we are ready to provision/update the disk encryption extensions
    # The following logic was mostly ported from xplat-cli
    public_config = {
        'KeyVaultURL': disk_encryption_keyvault_url,
        'VolumeType': volume_type,
        'EncryptionOperation': 'EnableEncryption' if not encrypt_format_all else 'EnableEncryptionFormatAll',
        'KeyEncryptionKeyURL': key_encryption_key,
        'KeyEncryptionAlgorithm': key_encryption_algorithm,
        'SequenceVersion': str(sequence_version),
    }
    if use_new_ade:
        public_config.update({
            "KeyVaultResourceId": disk_encryption_keyvault,
            "KekVaultResourceId": key_encryption_keyvault if key_encryption_key else '',
        })
    else:
        public_config.update({
            'AADClientID': aad_client_id,
            'AADClientCertThumbprint': aad_client_cert_thumbprint,
        })

    ade_legacy_private_config = {
        'AADClientSecret': aad_client_secret if is_linux else (aad_client_secret or '')
    }

    from .operations.vm_extension import VMExtensionCreate
    poller = VMExtensionCreate(cli_ctx=cmd.cli_ctx)(command_args={
        'resource_group': resource_group_name,
        'vm_name': vm_name,
        'vm_extension_name': extension['name'],
        'location': vm['location'],
        'publisher': extension['publisher'],
        'type': extension['name'],
        'protected_settings': None if use_new_ade else ade_legacy_private_config,
        'type_handler_version': extension['version'] if use_new_ade else extension['legacy_version'],
        'settings': public_config,
        'auto_upgrade_minor_version': True,
    })
    extension_result = LongRunningOperation(cmd.cli_ctx)(poller)

    # verify the extension was ok
    if extension_result['provisioningState'] != 'Succeeded':
        raise CLIError('Extension needed for disk encryption was not provisioned correctly')

    if not use_new_ade:
        if not extension_result.get('instanceView', {}).get('statuses', [{}])[0].get('message', ''):
            raise CLIError('Could not find url pointing to the secret for disk encryption')

        # 3. update VM's storage profile with the secrets
        status_url = extension_result['instanceView']['statuses'][0]['message']

        vm = VMShow(cli_ctx=cmd.cli_ctx)(command_args={
            'vm_name': vm_name,
            'resource_group': resource_group_name
        })

        settings = {
            "storageProfile.osDisk.encryptionSettings.diskEncryptionKey.secretUrl": status_url,
            "storageProfile.osDisk.encryptionSettings.diskEncryptionKey.sourceVault.id": disk_encryption_keyvault,
            "storageProfile.osDisk.encryptionSettings.enabled": "True"
        }
        if key_encryption_key:
            settings.update({
                "storageProfile.osDisk.encryptionSettings.keyEncryptionKey.keyUrl": key_encryption_key,
                "storageProfile.osDisk.encryptionSettings.keyEncryptionKey.sourceVault.id": key_encryption_keyvault
            })
        disk_encryption_settings = " ".join([f"{k}={v}" for k, v in settings.items()])

        if vm_encrypted:
            # stop the vm before update if the vm is already encrypted
            logger.warning("Deallocating the VM before updating encryption settings...")
            from .aaz.latest.vm import Deallocate as VMDeallocate
            VMDeallocate(cli_ctx=cmd.cli_ctx)(command_args={
                'vm_name': vm_name,
                'resource_group': resource_group_name
            })

        from .operations.vm import VMUpdate
        LongRunningOperation(cmd.cli_ctx)(
            VMUpdate(cli_ctx=cmd.cli_ctx)(command_args={
                'set': disk_encryption_settings
            })
        )

        if vm_encrypted:
            # and start after the update
            logger.warning("Restarting the VM after the update...")
            from .aaz.latest.vm import Start as VMStart
            VMStart(cli_ctx=cmd.cli_ctx)(command_args={
                'vm_name': vm_name,
                'resource_group': resource_group_name
            })

    if is_linux and volume_type != _DATA_VOLUME_TYPE:
        old_ade_msg = "If you see 'VMRestartPending', please restart the VM, and the encryption will finish shortly"
        logger.warning("The encryption request was accepted. Please use 'show' command to monitor "
                       "the progress. %s", "" if use_new_ade else old_ade_msg)


def decrypt_vm(cmd, resource_group_name, vm_name, volume_type=None, force=False):
    from knack.util import CLIError

    from .operations.vm import VMShow
    vm = VMShow(cli_ctx=cmd.cli_ctx)(command_args={
        'vm_name': vm_name,
        'resource_group': resource_group_name,
    })
    has_new_ade, has_old_ade = _detect_ade_status(vm)
    if not has_new_ade and not has_old_ade:
        logger.warning('Azure Disk Encryption is not enabled')
        return
    is_linux = _is_linux_os_aaz(vm)
    # pylint: disable=no-member

    # 1. be nice, figure out the default volume type and also verify VM will not be busted
    if is_linux:
        if volume_type:
            if not force and volume_type != _DATA_VOLUME_TYPE:
                raise CLIError("Only Data disks can have encryption disabled in a Linux VM. "
                               "Use '--force' to ignore the warning")
        else:
            volume_type = _DATA_VOLUME_TYPE
    elif volume_type is None:
        volume_type = _ALL_VOLUME_TYPE

    extension = vm_extension_info['Linux' if is_linux else 'Windows']
    # sequence_version should be incremented since encryptions occurred before
    sequence_version = uuid.uuid4()

    # 2. update the disk encryption extension
    # The following logic was mostly ported from xplat-cli
    public_config = {
        'VolumeType': volume_type,
        'EncryptionOperation': 'DisableEncryption',
        'SequenceVersion': str(sequence_version),
    }

    from .operations.vm_extension import VMExtensionCreate
    poller = VMExtensionCreate(cli_ctx=cmd.cli_ctx)(command_args={
        'resource_group': resource_group_name,
        'vm_name': vm_name,
        'vm_extension_name': extension['name'],
        'location': vm['location'],
        'publisher': extension['publisher'],
        'type': extension['name'] if has_new_ade else extension['legacy_version'],
        'settings': public_config,
        'auto_upgrade_minor_version': True,
    })
    extension_result = LongRunningOperation(cmd.cli_ctx)(poller)
    if extension_result['provisioningState'] != 'Succeeded':
        raise CLIError("Extension updating didn't succeed")

    if not has_new_ade:
        # 3. Remove the secret from VM's storage profile
        from .operations.vm import VMUpdate

        class RemoveSecret(VMUpdate):
            def pre_instance_update(self, instance):
                instance.properties.storage_profile.os_disk.encryption_settings = {'enabled': False}

        LongRunningOperation(cmd.cli_ctx)(
            RemoveSecret(cli_ctx=cmd.cli_ctx)(command_args={
                'vm_name': vm_name,
                'resource_group': resource_group_name
            })
        )


def _show_vm_encryption_status_thru_new_ade(vm_instance_view):
    ade = _find_existing_ade(vm_instance_view, use_instance_view=True)
    disk_infos = []
    for div in vm_instance_view.get('instanceView', {}).get('disks', []):
        disk_infos.append({
            'name': div['name'],
            'encryptionSettings': div['encryptionSettings'],
            'statuses': [x for x in div.get('statuses', []) if x.get('code', '').startswith('EncryptionState')],
        })

    return {
        'status': ade['statuses'] if ade else None,
        'substatus': ade.get('substatuses', None) if ade else None,
        'disks': disk_infos
    }


def show_vm_encryption_status(cmd, resource_group_name, vm_name):

    encryption_status = {
        'osDisk': 'NotEncrypted',
        'osDiskEncryptionSettings': None,
        'dataDisk': 'NotEncrypted',
        'osType': None
    }
    from .operations.vm import VMShow
    vm = VMShow(cli_ctx=cmd.cli_ctx)(command_args={
        'vm_name': vm_name,
        'resource_group': resource_group_name,
        'expand': 'instanceView'
    })
    has_new_ade, has_old_ade = _detect_ade_status(vm)
    if not has_new_ade and not has_old_ade:
        logger.warning('Azure Disk Encryption is not enabled')
        return None
    if has_new_ade:
        return _show_vm_encryption_status_thru_new_ade(vm)
    is_linux = _is_linux_os_aaz(vm)

    # pylint: disable=no-member
    # The following logic was mostly ported from xplat-cli
    os_type = 'Linux' if is_linux else 'Windows'
    encryption_status['osType'] = os_type
    extension = vm_extension_info[os_type]
    from .operations.vm_extension import VMExtensionShow
    extension_result = VMExtensionShow(cli_ctx=cmd.cli_ctx)(command_args={
        'vm_name': vm_name,
        'resource_group': resource_group_name,
        'name': extension['name'],
        'expand': 'instanceView'
    })
    logger.debug(extension_result)
    if _statuses := extension_result['instanceView'].get('statuses', []):
        encryption_status['progressMessage'] = _statuses[0].message

    substatus_message = None
    if _substatuses := extension_result['instanceView'].get('substatuses', []):
        substatus_message = _substatuses[0].message

    encryption_status['osDiskEncryptionSettings'] = vm['storageProfile'].get('osDisk', {}).get('encryptionSettings', None)

    import json
    if is_linux:
        try:
            message_object = json.loads(substatus_message)
        except Exception:  # pylint: disable=broad-except
            message_object = None  # might be from outdated extension

        if message_object and ('os' in message_object):
            encryption_status['osDisk'] = message_object['os']
        else:
            encryption_status['osDisk'] = 'Unknown'

        if message_object and 'data' in message_object:
            encryption_status['dataDisk'] = message_object['data']
        else:
            encryption_status['dataDisk'] = 'Unknown'
    else:
        # Windows - get os and data volume encryption state from the vm model
        if (encryption_status['osDiskEncryptionSettings'] and
                bool(encryption_status['osDiskEncryptionSettings'].get('enabled', '')) and
                encryption_status['osDiskEncryptionSettings'].get('diskEncryptionKey', {}).get('secret_url', '')):
            encryption_status['osDisk'] = _STATUS_ENCRYPTED
        else:
            encryption_status['osDisk'] = 'Unknown'

        if extension_result['provisioning_state'] == 'Succeeded':
            volume_type = extension_result.get('settings', {}).get('VolumeType', None)
            about_data_disk = not volume_type or volume_type.lower() != 'os'
            if about_data_disk and extension_result.get('settings', {}).get('EncryptionOperation', None) == 'EnableEncryption':
                encryption_status['dataDisk'] = _STATUS_ENCRYPTED

    return encryption_status


def _get_keyvault_key_url(cli_ctx, keyvault_name, key_name):
    vault_base_url = get_key_vault_base_url(cli_ctx, keyvault_name)
    client = create_data_plane_keyvault_key_client(cli_ctx, vault_base_url)
    key = client.get_key(key_name)
    return key.id


def _handles_default_volume_type_for_vmss_encryption(is_linux, volume_type, force):
    if is_linux:
        volume_type = volume_type or _DATA_VOLUME_TYPE
        if volume_type != _DATA_VOLUME_TYPE:
            msg = 'OS disk encyrption is not yet supported for Linux VM scale sets'
            if force:
                logger.warning(msg)
            else:
                from knack.util import CLIError
                raise CLIError(msg)
    else:
        volume_type = volume_type or _ALL_VOLUME_TYPE
    return volume_type


def encrypt_vmss(cmd, resource_group_name, vmss_name,  # pylint: disable=too-many-locals, too-many-statements
                 disk_encryption_keyvault,
                 key_encryption_keyvault=None,
                 key_encryption_key=None,
                 key_encryption_algorithm='RSA-OAEP',
                 volume_type=None,
                 encryption_identity=None,
                 force=False):
    from azure.mgmt.core.tools import parse_resource_id
    from knack.util import CLIError

    from .aaz.latest.vmss import Update as VMSSUpdate

    class VMSSEncrypt(VMSSUpdate):
        def pre_instance_update(self, instance):
            _disk_encryption_keyvault, _key_encryption_keyvault, _key_encryption_key, _volume_type = \
                disk_encryption_keyvault, key_encryption_keyvault, key_encryption_key, volume_type

            is_linux = _is_linux_os_aaz(instance.properties.virtual_machine_profile.to_serialized_data())
            extension = vm_extension_info['Linux' if is_linux else 'Windows']

            # 1. First validate arguments
            _volume_type = _handles_default_volume_type_for_vmss_encryption(is_linux, _volume_type, force)

            # retrieve keyvault details
            _disk_encryption_keyvault_url = get_key_vault_base_url(
                cmd.cli_ctx, (parse_resource_id(_disk_encryption_keyvault))['name']
            )

            # disk encryption key itself can be further protected, so let us verify
            if _key_encryption_key:
                _key_encryption_keyvault = _key_encryption_keyvault or _disk_encryption_keyvault

            if encryption_identity:
                result = updateVmssEncryptionSetting(cmd, instance.to_serialized_data(), resource_group_name, vmss_name, encryption_identity)
                if result:
                    logger.info("Encryption Identity successfully set in virtual machine scale set")
                else:
                    raise CLIError("Failed to update encryption Identity to the VMSS")

            #  to avoid bad server errors, ensure the vault has the right configurations
            _verify_keyvault_good_for_encryption(cmd.cli_ctx, _disk_encryption_keyvault, _key_encryption_keyvault, instance.to_serialized_data(), force)

            # if key name and not key url, get url.
            if _key_encryption_key and '://' not in _key_encryption_key:
                _key_encryption_key = _get_keyvault_key_url(
                    cmd.cli_ctx, (parse_resource_id(_key_encryption_keyvault))['name'], _key_encryption_key)

            # 2. we are ready to provision/update the disk encryption extensions
            public_config = {
                'KeyVaultURL': _disk_encryption_keyvault_url,
                'KeyEncryptionKeyURL': _key_encryption_key or '',
                "KeyVaultResourceId": _disk_encryption_keyvault,
                "KekVaultResourceId": _key_encryption_keyvault if _key_encryption_key else '',
                'KeyEncryptionAlgorithm': key_encryption_algorithm if _key_encryption_key else '',
                'VolumeType': _volume_type,
                'EncryptionOperation': 'EnableEncryption'
            }

            ext = {
                'name': extension['name'],
                'properties': {
                    'publisher': extension['publisher'],
                    'type_handler_version': extension['version'],
                    'type': extension['name'],
                    'settings': public_config,
                    'auto_upgrade_minor_version': True,
                    'force_update_tag': str(uuid.uuid4())
                }
            }
            exts = [ext]

            # remove any old ade extensions set by this command and add the new one.
            vmss_ext_profile = instance.properties.virtual_machine_profile.extension_profile
            if vmss_ext_profile and vmss_ext_profile.extensions:
                exts.extend(old_ext for old_ext in instance.properties.virtual_machine_profile.extension_profile.extensions
                            if old_ext.properties.type != ext['properties']['type'] or old_ext.name != ext['name'])
            instance.properties.virtual_machine_profile.extension_profile.extensions = exts

            # Avoid unnecessary permission error
            instance.properties.virtual_machine_profile.storage_profile.image_reference = None

        def post_operations(self):
            _show_post_action_message(resource_group_name, vmss_name, self.ctx.vars.instance.properties.upgrade_policy.mode == "Manual", True)

    poller = VMSSEncrypt(cli_ctx=cmd.cli_ctx)(command_args={
        'resource_group': resource_group_name,
        'vm_scale_set_name': vmss_name
    })
    LongRunningOperation(cmd.cli_ctx)(poller)


def decrypt_vmss(cmd, resource_group_name, vmss_name, volume_type=None, force=False):
    from .aaz.latest.vmss import Update

    class VMSSDecrypt(Update):
        def pre_instance_update(self, instance):
            is_linux = _is_linux_os_aaz(instance.properties.virtual_machine_profile.to_serialized_data())
            extension = vm_extension_info['Linux' if is_linux else 'Windows']

            # 1. be nice, figure out the default volume type
            _volume_type = _handles_default_volume_type_for_vmss_encryption(is_linux, volume_type, force)

            # 2. update the disk encryption extension
            public_config = {
                'VolumeType': _volume_type,
                'EncryptionOperation': 'DisableEncryption',
            }

            ext = {
                'name': extension['name'],
                'properties': {
                    'publisher': extension['publisher'],
                    'type_handler_version': extension['version'],
                    'type': extension['name'],
                    'settings': public_config,
                    'auto_upgrade_minor_version': True,
                    'force_update_tag': str(uuid.uuid4())
                }
            }

            ade_extension_matched = False
            for i, x in enumerate(instance.properties.virtual_machine_profile.extension_profile.extensions):
                if (str(x.name).lower() == extension['name'].lower() and
                        str(x.properties.publisher).lower() == extension['publisher'].lower()):
                    ade_extension_matched = True
                    instance.properties.virtual_machine_profile.extension_profile.extensions[i] = ext
                    break

            if not ade_extension_matched:
                from knack.util import CLIError
                raise CLIError("VM scale set '{}' was not encrypted".format(vmss_name))

            # Avoid unnecessary permission error
            instance.properties.virtual_machine_profile.storage_profile.image_reference = None

        def post_operations(self):
            _show_post_action_message(resource_group_name, vmss_name, self.ctx.vars.instance.properties.upgrade_policy.mode == "Manual", False)

    poller = VMSSDecrypt(cli_ctx=cmd.cli_ctx)(command_args={
        'resource_group': resource_group_name,
        'vm_scale_set_name': vmss_name
    })
    LongRunningOperation(cmd.cli_ctx)(poller)


def _show_post_action_message(resource_group_name, vmss_name, maunal_mode, enable):
    msg = ''
    if maunal_mode:
        msg = ("With manual upgrade mode, you will need to run 'az vmss update-instances -g {} -n {} "
               "--instance-ids \"*\"' to propagate the change.\n".format(resource_group_name, vmss_name))
    msg += ("Note, {} encryption will take a while to finish. Please query the status using "
            "'az vmss encryption show -g {} -n {}'. For Linux VM, you will lose the access during the period".format(
                'enabling' if enable else 'disabling', resource_group_name, vmss_name))
    logger.warning(msg)


def show_vmss_encryption_status(cmd, resource_group_name, vmss_name):
    from .operations.vmss import VMSSListInstances
    vm_instances = VMSSListInstances(cli_ctx=cmd.cli_ctx)(command_args={
        'virtual_machine_scale_set_name': vmss_name,
        'resource_group': resource_group_name,
        'expand': 'instanceView',
        'select': 'instanceView'
    })

    result = []
    for instance in vm_instances:
        view = instance['instanceView']
        disk_infos = []
        vm_enc_info = {
            'id': instance['id'],
            'disks': disk_infos
        }
        for div in view['disks']:
            disk_infos.append({
                'name': div['name'],
                'encryptionSettings': div.get('encryptionSettings', []),
                'statuses': [x for x in (div.get('statuses', [])) if (x.get('code', '')).startswith('EncryptionState')]
            })

        result.append(vm_enc_info)
    return result


# aaz implementation
def _verify_keyvault_good_for_encryption(cli_ctx, disk_vault_id, key_vault_id, vm_or_vmss, force):
    def _report_client_side_validation_error(msg):
        if force:
            logger.warning("WARNING: %s %s", msg, "Encryption might fail.")
        else:
            from knack.util import CLIError
            raise CLIError("ERROR: {}".format(msg))

    resource_type = "VMSS" if vm_or_vmss['type'].lower().endswith("virtualmachinescalesets") else "VM"

    from azure.cli.core.commands.client_factory import get_mgmt_service_client
    from azure.cli.core.profiles import ResourceType
    from azure.mgmt.core.tools import parse_resource_id

    client = get_mgmt_service_client(cli_ctx, ResourceType.MGMT_KEYVAULT).vaults
    disk_vault_resource_info = parse_resource_id(disk_vault_id)
    key_vault = client.get(disk_vault_resource_info['resource_group'], disk_vault_resource_info['name'])

    # ensure vault has 'EnabledForDiskEncryption' permission or VM has encryption identity set for ADE operation
    if resource_type == 'VM':
        vm_encryption_identity = vm_or_vmss
    else:
        vm_encryption_identity = vm_or_vmss['properties']['virtualMachineProfile']

    if vm_encryption_identity.get('securityProfile', {}).get('encryptionIdentity', {}).get('userAssignedIdentityResourceId', None):
        pass
    elif not key_vault.properties or not key_vault.properties.enabled_for_disk_encryption:
        _report_client_side_validation_error(
            "Keyvault '{}' is not enabled for disk encryption.".format(disk_vault_resource_info['resource_name']))

    if key_vault_id:
        kek_vault_info = parse_resource_id(key_vault_id)
        if disk_vault_resource_info['name'].lower() != kek_vault_info['name'].lower():
            client.get(kek_vault_info['resource_group'], kek_vault_info['name'])

    # verify subscription mataches
    vm_vmss_resource_info = parse_resource_id(vm_or_vmss['id'])
    if vm_vmss_resource_info['subscription'].lower() != disk_vault_resource_info['subscription'].lower():
        _report_client_side_validation_error("{} {}'s subscription does not match keyvault's subscription."
                                             .format(resource_type, vm_vmss_resource_info['name']))

    # verify region matches
    if key_vault.location.replace(' ', '').lower() != vm_or_vmss['location'].replace(' ', '').lower():
        _report_client_side_validation_error(
            "{} {}'s region does not match keyvault's region.".format(resource_type, vm_vmss_resource_info['name']))



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
#
# Code generated by aaz-dev-tools
# --------------------------------------------------------------------------------------------

# pylint: skip-file
# flake8: noqa

from azure.cli.core.aaz import *


@register_command(
    "network private-dns zone delete",
    confirmation="Are you sure you want to perform this operation?",
)
class Delete(AAZCommand):
    """Delete a Private DNS zone.

    All DNS records in the zone will also be deleted. This operation cannot be undone. Private DNS zone cannot be deleted unless all virtual network links to it are removed.

    :example: Delete a Private DNS zone using a fully qualified domain name.
        az network private-dns zone delete -g MyResourceGroup -n www.mysite.com
    """

    _aaz_info = {
        "version": "2018-09-01",
        "resources": [
            ["mgmt-plane", "/subscriptions/{}/resourcegroups/{}/providers/microsoft.network/privatednszones/{}", "2018-09-01"],
        ]
    }

    AZ_SUPPORT_NO_WAIT = True

    def _handler(self, command_args):
        super()._handler(command_args)
        return self.build_lro_poller(self._execute_operations, None)

    _args_schema = None

    @classmethod
    def _build_arguments_schema(cls, *args, **kwargs):
        if cls._args_schema is not None:
            return cls._args_schema
        cls._args_schema = super()._build_arguments_schema(*args, **kwargs)

        # define Arg Group ""

        _args_schema = cls._args_schema
        _args_schema.if_match = AAZStrArg(
            options=["--if-match"],
            help="ETag of the Private DNS zone. Omit this value to always delete the current zone. Specify the last-seen ETag value to prevent accidentally deleting any concurrent changes.",
        )
        _args_schema.name = AAZStrArg(
            options=["-n", "--name"],
            help="Name of the Private DNS zone.",
            required=True,
            id_part="name",
        )
        _args_schema.resource_group = AAZResourceGroupNameArg(
            required=True,
        )
        return cls._args_schema

    def _execute_operations(self):
        self.pre_operations()
        yield self.PrivateZonesDelete(ctx=self.ctx)()
        self.post_operations()

    @register_callback
    def pre_operations(self):
        pass

    @register_callback
    def post_operations(self):
        pass

    class PrivateZonesDelete(AAZHttpOperation):
        CLIENT_TYPE = "MgmtClient"

        def __call__(self, *args, **kwargs):
            request = self.make_request()
            session = self.client.send_request(request=request, stream=False, **kwargs)
            if session.http_response.status_code in [202]:
                return self.client.build_lro_polling(
                    self.ctx.args.no_wait,
                    session,
                    self.on_200,
                    self.on_error,
                    lro_options={"final-state-via": "azure-async-operation"},
                    path_format_arguments=self.url_parameters,
                )
            if session.http_response.status_code in [200]:
                return self.client.build_lro_polling(
                    self.ctx.args.no_wait,
                    session,
                    self.on_200,
                    self.on_error,
                    lro_options={"final-state-via": "azure-async-operation"},
                    path_format_arguments=self.url_parameters,
                )
            if session.http_response.status_code in [204]:
                return self.client.build_lro_polling(
                    self.ctx.args.no_wait,
                    session,
                    self.on_204,
                    self.on_error,
                    lro_options={"final-state-via": "azure-async-operation"},
                    path_format_arguments=self.url_parameters,
                )

            return self.on_error(session.http_response)

        @property
        def url(self):
            return self.client.format_url(
                "/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/Microsoft.Network/privateDnsZones/{privateZoneName}",
                **self.url_parameters
            )

        @property
        def method(self):
            return "DELETE"

        @property
        def error_format(self):
            return "ODataV4Format"

        @property
        def url_parameters(self):
            parameters = {
                **self.serialize_url_param(
                    "privateZoneName", self.ctx.args.name,
                    required=True,
                ),
                **self.serialize_url_param(
                    "resourceGroupName", self.ctx.args.resource_group,
                    required=True,
                ),
                **self.serialize_url_param(
                    "subscriptionId", self.ctx.subscription_id,
                    required=True,
                ),
            }
            return parameters

        @property
        def query_parameters(self):
            parameters = {
                **self.serialize_query_param(
                    "api-version", "2018-09-01",
                    required=True,
                ),
            }
            return parameters

        @property
        def header_parameters(self):
            parameters = {
                **self.serialize_header_param(
                    "If-Match", self.ctx.args.if_match,
                ),
            }
            return parameters

        def on_200(self, session):
            pass

        def on_204(self, session):
            pass


class _DeleteHelper:
    """Helper class for Delete"""


__all__ = ["Delete"]



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
#
# Code generated by aaz-dev-tools
# --------------------------------------------------------------------------------------------

# pylint: skip-file
# flake8: noqa

from azure.cli.core.aaz import *


@register_command(
    "security contact create",
    is_preview=True,
)
class Create(AAZCommand):
    """Creates a security contact.

    :example: Creates a security contact.
        az security contact create -n 'default' --emails 'john@contoso.com;jane@contoso.com' --phone '214-275-4038' --notifications-by-role '{"state":"On","roles":["Owner"]}' --alert-notifications '{"state":"On","minimalSeverity":"Low"}'
    """

    _aaz_info = {
        "version": "2020-01-01-preview",
        "resources": [
            ["mgmt-plane", "/subscriptions/{}/providers/microsoft.security/securitycontacts/{}", "2020-01-01-preview"],
        ]
    }

    def _handler(self, command_args):
        super()._handler(command_args)
        self._execute_operations()
        return self._output()

    _args_schema = None

    @classmethod
    def _build_arguments_schema(cls, *args, **kwargs):
        if cls._args_schema is not None:
            return cls._args_schema
        cls._args_schema = super()._build_arguments_schema(*args, **kwargs)

        # define Arg Group ""

        _args_schema = cls._args_schema
        _args_schema.security_contact_name = AAZStrArg(
            options=["-n", "--name", "--security-contact-name"],
            help="Name of the security contact object",
            required=True,
        )

        # define Arg Group "Properties"

        _args_schema = cls._args_schema
        _args_schema.alert_notifications = AAZObjectArg(
            options=["--alert-notifications"],
            arg_group="Properties",
            help="Defines whether to send email notifications about new security alerts",
        )
        _args_schema.emails = AAZStrArg(
            options=["--emails"],
            arg_group="Properties",
            help="List of email addresses which will get notifications from Microsoft Defender for Cloud by the configurations defined in this security contact.",
        )
        _args_schema.notifications_by_role = AAZObjectArg(
            options=["--notifications-by-role"],
            arg_group="Properties",
            help="Defines whether to send email notifications from Microsoft Defender for Cloud to persons with specific RBAC roles on the subscription.",
        )
        _args_schema.phone = AAZStrArg(
            options=["--phone"],
            arg_group="Properties",
            help="The security contact's phone number",
        )

        alert_notifications = cls._args_schema.alert_notifications
        alert_notifications.minimal_severity = AAZStrArg(
            options=["minimal-severity"],
            help="Defines the minimal alert severity which will be sent as email notifications",
            enum={"High": "High", "Low": "Low", "Medium": "Medium"},
        )
        alert_notifications.state = AAZStrArg(
            options=["state"],
            help="Defines if email notifications will be sent about new security alerts",
            enum={"Off": "Off", "On": "On"},
        )

        notifications_by_role = cls._args_schema.notifications_by_role
        notifications_by_role.roles = AAZListArg(
            options=["roles"],
            help="Defines which RBAC roles will get email notifications from Microsoft Defender for Cloud. List of allowed RBAC roles: ",
        )
        notifications_by_role.state = AAZStrArg(
            options=["state"],
            help="Defines whether to send email notifications from AMicrosoft Defender for Cloud to persons with specific RBAC roles on the subscription.",
            enum={"Off": "Off", "On": "On"},
        )

        roles = cls._args_schema.notifications_by_role.roles
        roles.Element = AAZStrArg(
            enum={"AccountAdmin": "AccountAdmin", "Contributor": "Contributor", "Owner": "Owner", "ServiceAdmin": "ServiceAdmin"},
        )
        return cls._args_schema

    def _execute_operations(self):
        self.pre_operations()
        self.SecurityContactsCreate(ctx=self.ctx)()
        self.post_operations()

    @register_callback
    def pre_operations(self):
        pass

    @register_callback
    def post_operations(self):
        pass

    def _output(self, *args, **kwargs):
        result = self.deserialize_output(self.ctx.vars.instance, client_flatten=True)
        return result

    class SecurityContactsCreate(AAZHttpOperation):
        CLIENT_TYPE = "MgmtClient"

        def __call__(self, *args, **kwargs):
            request = self.make_request()
            session = self.client.send_request(request=request, stream=False, **kwargs)
            if session.http_response.status_code in [200, 201]:
                return self.on_200_201(session)

            return self.on_error(session.http_response)

        @property
        def url(self):
            return self.client.format_url(
                "/subscriptions/{subscriptionId}/providers/Microsoft.Security/securityContacts/{securityContactName}",
                **self.url_parameters
            )

        @property
        def method(self):
            return "PUT"

        @property
        def error_format(self):
            return "MgmtErrorFormat"

        @property
        def url_parameters(self):
            parameters = {
                **self.serialize_url_param(
                    "securityContactName", self.ctx.args.security_contact_name,
                    required=True,
                ),
                **self.serialize_url_param(
                    "subscriptionId", self.ctx.subscription_id,
                    required=True,
                ),
            }
            return parameters

        @property
        def query_parameters(self):
            parameters = {
                **self.serialize_query_param(
                    "api-version", "2020-01-01-preview",
                    required=True,
                ),
            }
            return parameters

        @property
        def header_parameters(self):
            parameters = {
                **self.serialize_header_param(
                    "Content-Type", "application/json",
                ),
                **self.serialize_header_param(
                    "Accept", "application/json",
                ),
            }
            return parameters

        @property
        def content(self):
            _content_value, _builder = self.new_content_builder(
                self.ctx.args,
                typ=AAZObjectType,
                typ_kwargs={"flags": {"required": True, "client_flatten": True}}
            )
            _builder.set_prop("properties", AAZObjectType, typ_kwargs={"flags": {"client_flatten": True}})

            properties = _builder.get(".properties")
            if properties is not None:
                properties.set_prop("alertNotifications", AAZObjectType, ".alert_notifications")
                properties.set_prop("emails", AAZStrType, ".emails")
                properties.set_prop("notificationsByRole", AAZObjectType, ".notifications_by_role")
                properties.set_prop("phone", AAZStrType, ".phone")

            alert_notifications = _builder.get(".properties.alertNotifications")
            if alert_notifications is not None:
                alert_notifications.set_prop("minimalSeverity", AAZStrType, ".minimal_severity")
                alert_notifications.set_prop("state", AAZStrType, ".state")

            notifications_by_role = _builder.get(".properties.notificationsByRole")
            if notifications_by_role is not None:
                notifications_by_role.set_prop("roles", AAZListType, ".roles")
                notifications_by_role.set_prop("state", AAZStrType, ".state")

            roles = _builder.get(".properties.notificationsByRole.roles")
            if roles is not None:
                roles.set_elements(AAZStrType, ".")

            return self.serialize_content(_content_value)

        def on_200_201(self, session):
            data = self.deserialize_http_content(session)
            self.ctx.set_var(
                "instance",
                data,
                schema_builder=self._build_schema_on_200_201
            )

        _schema_on_200_201 = None

        @classmethod
        def _build_schema_on_200_201(cls):
            if cls._schema_on_200_201 is not None:
                return cls._schema_on_200_201

            cls._schema_on_200_201 = AAZObjectType()

            _schema_on_200_201 = cls._schema_on_200_201
            _schema_on_200_201.id = AAZStrType(
                flags={"read_only": True},
            )
            _schema_on_200_201.name = AAZStrType(
                flags={"read_only": True},
            )
            _schema_on_200_201.properties = AAZObjectType(
                flags={"client_flatten": True},
            )
            _schema_on_200_201.type = AAZStrType(
                flags={"read_only": True},
            )

            properties = cls._schema_on_200_201.properties
            properties.alert_notifications = AAZObjectType(
                serialized_name="alertNotifications",
            )
            properties.emails = AAZStrType()
            properties.notifications_by_role = AAZObjectType(
                serialized_name="notificationsByRole",
            )
            properties.phone = AAZStrType()

            alert_notifications = cls._schema_on_200_201.properties.alert_notifications
            alert_notifications.minimal_severity = AAZStrType(
                serialized_name="minimalSeverity",
            )
            alert_notifications.state = AAZStrType()

            notifications_by_role = cls._schema_on_200_201.properties.notifications_by_role
            notifications_by_role.roles = AAZListType()
            notifications_by_role.state = AAZStrType()

            roles = cls._schema_on_200_201.properties.notifications_by_role.roles
            roles.Element = AAZStrType()

            return cls._schema_on_200_201


class _CreateHelper:
    """Helper class for Create"""


__all__ = ["Create"]



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from knack.log import get_logger
from ._resource_config import (
    RESOURCE,
    AUTH_TYPE
)
from ._validators import (
    get_source_resource_name,
    get_target_resource_name,
)

logger = get_logger(__name__)

AUTHTYPES = {
    AUTH_TYPE.SystemIdentity: 'systemAssignedIdentity',
    AUTH_TYPE.UserIdentity: 'userAssignedIdentity',
    AUTH_TYPE.ServicePrincipalSecret: 'servicePrincipalSecret',
    AUTH_TYPE.UserAccount: 'userAccount',
    AUTH_TYPE.WorkloadIdentity: 'userAssignedIdentity'
}


# pylint: disable=line-too-long, consider-using-f-string
# For db(mysqlFlex/psql/psqlFlex/sql) linker with AAD auth type, it's passwordless connection and requires extension installation
def is_passwordless_command(cmd, auth_info):
    # return if connection is not for db mi
    if auth_info['auth_type'] not in [AUTHTYPES[AUTH_TYPE.SystemIdentity],
                                      AUTHTYPES[AUTH_TYPE.UserAccount],
                                      AUTHTYPES[AUTH_TYPE.ServicePrincipalSecret],
                                      AUTHTYPES[AUTH_TYPE.UserIdentity]]:
        return False
    source_type = get_source_resource_name(cmd)
    target_type = get_target_resource_name(cmd)
    if source_type not in {RESOURCE.WebApp, RESOURCE.ContainerApp, RESOURCE.SpringCloud, RESOURCE.SpringCloudDeprecated, RESOURCE.FunctionApp, RESOURCE.Local}:
        return False
    if target_type not in {RESOURCE.Sql, RESOURCE.Postgres, RESOURCE.PostgresFlexible, RESOURCE.MysqlFlexible, RESOURCE.FabricSql}:
        return False
    return True



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

# pylint: disable=too-few-public-methods
from azure.cli.core.auth.util import resource_to_scopes


# This class is used to pass in custom token audience that will be respected by the SDK.
# Users can configure an audience based on their cloud.
class AppConfigurationCliCredential:

    def __init__(self, credential, resource: str = None):
        self._impl = credential
        self._resource = resource

    def get_token(self, *scopes, **kwargs):

        if self._resource is not None:
            scopes = resource_to_scopes(self._resource)

        return self._impl.get_token(*scopes, **kwargs)
